from typing import Optional, Any, Dict, List

from botocore.exceptions import ClientError

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.iam.role import IamRole
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class GlueCrawlerResource(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/glue.html
    """

    resource_type = "GlueCrawlerResource"
    service_name = "glue"

    # Name of the crawler.
    name: str
    # The IAM role for the crawler
    iam_role: IamRole
    # The Glue database where results are written,
    # such as: arn:aws:daylight:us-east-1::database/sometable/* .
    database_name: Optional[str] = None
    # A description of the new crawler.
    description: Optional[str] = None
    # A list of collection of targets to crawl.
    targets: Optional[Dict[str, List[dict]]] = None
    # A cron expression used to specify the schedule
    # For example, to run something every day at 12:15 UTC,
    # you would specify: cron(15 12 * * ? *) .
    schedule: Optional[str] = None
    # A list of custom classifiers that the user has registered.
    # By default, all built-in classifiers are included in a crawl,
    # but these custom classifiers always override the default classifiers for a given classification.
    classifiers: Optional[List[str]] = None
    # The table prefix used for catalog tables that are created.
    table_prefix: Optional[str] = None
    # The policy for the crawler's update and deletion behavior.
    schema_change_policy: Optional[Dict[str, str]] = None
    # A policy that specifies whether to crawl the entire dataset again,
    # or to crawl only folders that were added since the last crawler run.
    recrawl_policy: Optional[Dict[str, str]] = None
    lineage_configuration: Optional[Dict[str, str]] = None
    lake_formation_configuration: Optional[Dict[str, str]] = None
    # Crawler configuration information. This versioned JSON string
    # allows users to specify aspects of a crawler's behavior.
    configuration: Optional[str] = None
    # The name of the SecurityConfiguration structure to be used by this crawler.
    crawler_security_configuration: Optional[str] = None
    # The tags to use with this crawler request.
    tags: Optional[Dict[str, str]] = None

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the GlueCrawlerResource

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # create a dict of args which are not null, otherwise aws type validation fails
            non_null_args = {}
            if self.database_name:
                non_null_args["DatabaseName"] = self.database_name
            if self.description:
                non_null_args["Description"] = self.description
            if self.targets:
                non_null_args["Targets"] = self.targets
            if self.schedule:
                non_null_args["Schedule"] = self.schedule
            if self.classifiers:
                non_null_args["Classifiers"] = self.classifiers
            if self.table_prefix:
                non_null_args["TablePrefix"] = self.table_prefix
            if self.schema_change_policy:
                non_null_args["SchemaChangePolicy"] = self.schema_change_policy
            if self.recrawl_policy:
                non_null_args["RecrawlPolicy"] = self.recrawl_policy
            if self.lineage_configuration:
                non_null_args["LineageConfiguration"] = self.lineage_configuration
            if self.lake_formation_configuration:
                non_null_args[
                    "LakeFormationConfiguration"
                ] = self.lake_formation_configuration
            if self.configuration:
                non_null_args["Configuration"] = self.configuration
            if self.crawler_security_configuration:
                non_null_args[
                    "CrawlerSecurityConfiguration"
                ] = self.crawler_security_configuration
            if self.tags:
                non_null_args["Tags"] = self.tags

            ## Create crawler
            # Get the service_client
            service_client = self.get_service_client(aws_client)
            # logger.debug(f"ServiceClient: {service_client}")
            # logger.debug(f"ServiceClient type: {type(service_client)}")
            try:
                print_info(f"Creating GlueCrawlerResource: {self.name}")
                iam_role_arn = self.iam_role.get_arn(aws_client)
                create_crawler_response = service_client.create_crawler(
                    Name=self.name,
                    Role=iam_role_arn,
                    **non_null_args,
                )
                # logger.debug(f"GlueCrawlerResource: {create_crawler_response}")
                # logger.debug(f"GlueCrawlerResource type: {type(create_crawler_response)}")

                if create_crawler_response is not None:
                    print_info(f"GlueCrawlerResource created: {self.name}")
                    self.active_resource = create_crawler_response
                    return True
            except Exception as e:
                print_error(
                    "GlueCrawlerResource could not be created, this operation is known to be buggy."
                )
                print_error("Please try again.")
                print_error(e)
                return False
        except Exception as e:
            print_error(
                f"Received error while creating {self.get_resource_type()}, this operation is known to be buggy."
            )
            print_error("Please try again.")
            print_error(e)
        return False

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the GlueCrawlerResource

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_client = self.get_service_client(aws_client)
            get_crawler_response = service_client.get_crawler(Name=self.name)
            # logger.debug(f"GlueCrawlerResource: {get_crawler_response}")
            # logger.debug(f"GlueCrawlerResource type: {type(get_crawler_response)}")

            creation_time = get_crawler_response.get("Crawler", {}).get(
                "CreationTime", None
            )
            last_crawl = get_crawler_response.get("Crawler", {}).get("LastCrawl", None)
            logger.debug(f"GlueCrawlerResource creation_time: {creation_time}")
            # logger.debug(f"GlueCrawlerResource last_crawl: {last_crawl}")
            if creation_time is not None:
                logger.debug(f"GlueCrawlerResource found: {self.name}")
                self.active_resource = get_crawler_response
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
            pass
        except Exception as e:
            print_error(f"Received error while reading {self.get_resource_type()}.")
            print_error("Please try again.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the GlueCrawlerResource

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # Delete the GlueCrawlerResource
            service_client = self.get_service_client(aws_client)
            self.active_resource = None
            delete_crawler_response = service_client.delete_crawler(Name=self.name)
            logger.debug(f"GlueCrawlerResource: {delete_crawler_response}")
            logger.debug(f"GlueCrawlerResource type: {type(delete_crawler_response)}")
            print_info(f"GlueCrawlerResource deleted: {self.name}")
            return True
        except Exception as e:
            print_error(
                f"Received error while deleting {self.get_resource_type()}, this operation is known to be buggy."
            )
            print_error("Please try again or delete resources manually.")
            print_error(e)
        return False

    def start_crawler(self, aws_client: AwsApiClient) -> bool:
        """Runs the GlueCrawlerResource

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        logger.debug(
            "Running {} | {}".format(self.get_resource_type(), self.get_resource_name())
        )
        try:
            # Get the service_client
            service_client = self.get_service_client(aws_client)
            # logger.debug(f"ServiceClient: {service_client}")
            # logger.debug(f"ServiceClient type: {type(service_client)}")
            start_crawler_response = service_client.start_crawler(Name=self.name)
            # logger.debug(f"start_crawler_response: {start_crawler_response}")

            if start_crawler_response is not None:
                return True
        except Exception as e:
            print_error("GlueCrawler could not be started")
            print_error("Please try again.")
            print_error(e)
        return False



from typing import Optional, Any, Dict, Union, List

from docker.models.volumes import Volume
from docker.errors import NotFound

from phidata.infra.docker.api_client import DockerApiClient
from phidata.infra.docker.resource.base import DockerResource
from phidata.utils.log import logger


class DockerVolume(DockerResource):
    resource_type: str = "Volume"

    # driver (str) – Name of the driver used to create the volume
    driver: Optional[str] = None
    # driver_opts (dict) – Driver options as a key-value dictionary
    driver_opts: Optional[Dict[str, Any]] = None
    # labels (dict) – Labels to set on the volume
    labels: Optional[Dict[str, Any]] = None

    def _create(self, docker_client: DockerApiClient) -> bool:
        """Creates the Volume on docker

        Args:
            docker_client: The DockerApiClient for the current cluster
        """

        logger.debug("Creating: {}".format(self.get_resource_name()))
        volume_name: Optional[str] = self.name
        volume_object: Optional[Volume] = None

        try:
            volume_object = docker_client.api_client.volumes.create(
                name=volume_name,
                driver=self.driver,
                driver_opts=self.driver_opts,
                labels=self.labels,
            )
            if volume_object is not None:
                self.verbose_log("Volume Created: {}".format(volume_object.name))
            else:
                self.verbose_log("Volume could not be created")
            # self.verbose_log("Volume {}".format(volume_object.attrs))
        except Exception as e:
            raise

        # By this step the volume should be created
        # Get the data from the volume object
        self.verbose_log("Validating volume is created")
        if volume_object is not None:
            _id: str = volume_object.id
            _short_id: str = volume_object.short_id
            _name: str = volume_object.name
            _attrs: str = volume_object.attrs
            if _id:
                self.verbose_log("_id: {}".format(_id))
                self.id = _id
            if _short_id:
                self.verbose_log("_short_id: {}".format(_short_id))
                self.short_id = _short_id
            if _name:
                self.verbose_log("_name: {}".format(_name))
            if _attrs:
                self.verbose_log("_attrs: {}".format(_attrs))
                # TODO: use json_to_dict(_attrs)
                self.attrs = _attrs  # type: ignore

            # TODO: Validate that the volume object is created properly
            self.active_resource = volume_object
            self.active_resource_class = Volume
            return True
        return False

    def _read(self, docker_client: DockerApiClient) -> Any:
        """Returns a Volume object if the volume is active on the docker_client"""

        logger.debug("Reading: {}".format(self.get_resource_name()))
        volume_name: Optional[str] = self.name

        try:
            volume_list: Optional[
                List[Volume]
            ] = docker_client.api_client.volumes.list()
            # self.verbose_log("volume_list: {}".format(volume_list))
            if volume_list is not None:
                for volume in volume_list:
                    if volume.name == volume_name:
                        self.verbose_log(f"Volume {volume_name} exists")
                        self.active_resource = volume
                        self.active_resource_class = Volume
                        return volume
        except Exception:
            self.verbose_log(f"Volume {volume_name} not found")

        return None

    def _delete(self, docker_client: DockerApiClient) -> bool:
        """Deletes the Volume on docker

        Args:
            docker_client: The DockerApiClient for the current cluster
        """

        logger.debug("Deleting: {}".format(self.get_resource_name()))
        volume_name: Optional[str] = self.name
        volume_object: Optional[Volume] = self._read(docker_client)
        # Return True if there is no Volume to delete
        if volume_object is None:
            return True

        # Delete Volume
        try:
            self.active_resource = None
            volume_object.remove(force=True)
        except Exception as e:
            logger.exception("Error while deleting volume: {}".format(e))

        # Validate that the volume is deleted
        self.verbose_log("Validating volume is deleted")
        try:
            self.verbose_log("Reloading volume_object: {}".format(volume_object))
            volume_object.reload()
        except NotFound as e:
            self.verbose_log("Got NotFound Exception, Volume is deleted")
            return True

        return False



import json

from typing import List, Optional
from phidata.infra.aws.resource.iam.role import IamRole, IamPolicy
from phidata.infra.aws.resource.s3.bucket import S3Bucket


def create_glue_iam_role(
    name: str,
    s3_buckets: Optional[List[S3Bucket]] = None,
    skip_create: bool = False,
    skip_delete: bool = False,
) -> IamRole:

    return IamRole(
        name=name,
        assume_role_policy_document=json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"Service": "glue.amazonaws.com"},
                        "Action": "sts:AssumeRole",
                    }
                ],
            }
        ),
        policies=[
            IamPolicy(
                name=f"glueS3CrawlerPolicy",
                policy_document=json.dumps(
                    {
                        "Version": "2012-10-17",
                        "Statement": [
                            {
                                "Effect": "Allow",
                                "Action": ["s3:ListBucket"],
                                "Resource": [
                                    f"arn:aws:s3:::{bucket.name}"
                                    for bucket in s3_buckets
                                ],
                            },
                            {
                                "Effect": "Allow",
                                "Action": ["s3:GetObject"],
                                "Resource": [
                                    f"arn:aws:s3:::{bucket.name}/*"
                                    for bucket in s3_buckets
                                ],
                            },
                        ],
                    }
                ),
            )
        ],
        policy_arns=["arn:aws:iam::aws:policy/service-role/AWSGlueServiceRole"],
        skip_create=skip_create,
        skip_delete=skip_delete,
    )



from typing import Optional

from phidata.asset.aws.athena.query import AthenaQuery
from phidata.utils.cli_console import print_info, print_heading
from phidata.utils.log import logger
from phidata.workflow.py.python_workflow import (
    PythonWorkflow,
    PythonWorkflowArgs,
)


class RunAthenaQueryArgs(PythonWorkflowArgs):
    query: AthenaQuery
    get_results: bool = False


class RunAthenaQuery(PythonWorkflow):
    def __init__(
        self,
        query: AthenaQuery,
        get_results: bool = False,
        name: str = "run_athena_query",
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ):
        super().__init__()
        try:
            self.args: RunAthenaQueryArgs = RunAthenaQueryArgs(
                query=query,
                get_results=get_results,
                name=name,
                task_id=task_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                entrypoint=run_athena_query,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def query(self) -> AthenaQuery:
        return self.args.query

    @query.setter
    def query(self, query: AthenaQuery) -> None:
        if query is not None:
            self.args.query = query

    @property
    def get_results(self) -> bool:
        return self.args.get_results

    @get_results.setter
    def get_results(self, get_results: bool) -> None:
        if get_results is not None:
            self.args.get_results = get_results


def run_athena_query(**kwargs) -> bool:

    args: RunAthenaQueryArgs = RunAthenaQueryArgs(**kwargs)
    # logger.debug("RunAthenaQueryArgs: {}".format(args))

    start_success = args.query.start_query()
    if start_success:
        logger.info("AthenaQuery started")
        if args.get_results:
            logger.info(f"Getting results")
            results = args.query.get_results()
            if results:
                print_heading("\nAthenaQuery result")
                print_info(results)
            return start_success
    return start_success



from pathlib import Path
from textwrap import dedent
from typing import Optional, Any, Dict, List, Union

from botocore.exceptions import ClientError

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.iam.role import IamRole
from phidata.infra.aws.resource.cloudformation.stack import CloudFormationStack
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class EksCluster(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/eks.html
    """

    resource_type = "EksCluster"
    service_name = "eks"

    # The unique name to give to your cluster.
    name: str
    # version: The desired Kubernetes version for your cluster.
    # If you don't specify a value here, the latest version available in Amazon EKS is used.
    version: Optional[str] = None

    # role: The IAM role that provides permissions for the Kubernetes control plane to make calls
    # to Amazon Web Services API operations on your behalf.
    # If role is None, a default role is created using role_name
    role: Optional[IamRole] = None
    # Name for the default role when role is None, use "name-role" if not provided
    role_name: Optional[str] = None

    # resources_vpc_config: The VPC configuration that's used by the cluster control plane.
    # Amazon EKS VPC resources have specific requirements to work properly with Kubernetes.
    # You must specify at least two subnets. You can specify up to five security groups.
    # If resources_vpc_config is None, a default CloudFormationStack is created and
    # the vpc_config from that stack is used.
    resources_vpc_config: Optional[Dict[str, Union[List[str], bool]]] = None
    # The CloudFormationStack to build resources_vpc_config if provided
    vpc_stack: Optional[CloudFormationStack] = None
    # If resources_vpc_config and vpc_stack are None
    # create a default CloudFormationStack using vpc_stack_name, use "name-vpc-stack" if vpc_stack_name is None
    vpc_stack_name: Optional[str] = None
    vpc_stack_template_url: str = "https://amazon-eks.s3.us-west-2.amazonaws.com/cloudformation/2020-10-29/amazon-eks-vpc-private-subnets.yaml"

    # The Kubernetes network configuration for the cluster.
    kubernetes_network_config: Optional[Dict[str, str]] = None
    # Enable or disable exporting the Kubernetes control plane logs for your cluster to CloudWatch Logs.
    # By default, cluster control plane logs aren't exported to CloudWatch Logs.
    logging: Optional[Dict[str, List[dict]]] = None
    # Unique, case-sensitive identifier that you provide to ensure the idempotency of the request.
    client_request_token: Optional[str] = None
    # The metadata to apply to the cluster to assist with categorization and organization.
    # Each tag consists of a key and an optional value. You define both.
    tags: Optional[Dict[str, str]] = None
    # The encryption configuration for the cluster.
    encryption_config: Optional[
        List[Dict[str, Union[List[str], Dict[str, str]]]]
    ] = None

    skip_delete = False
    # bump the wait time for Eks to 30 seconds
    waiter_delay = 30
    # If True, updates the kubeconfig
    update_kubeconfig: bool = True
    # The kubeconfig_path to update
    kubeconfig_path: Path = Path.home().joinpath(".kube").joinpath("config").resolve()
    kubeconfig_cluster_name: Optional[str] = None
    kubeconfig_cluster_user: Optional[str] = None
    kubeconfig_cluster_context: Optional[str] = None

    # provided by api on create
    created_at: Optional[str] = None
    cluster_status: Optional[str] = None

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the EksCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # Create the IamRole
            eks_iam_role = self.get_eks_iam_role()
            eks_iam_role_arn: Optional[str] = None
            try:
                eks_iam_role.create(aws_client)
                eks_iam_role_arn = eks_iam_role.read(aws_client).arn
                print_info(f"ARN for {eks_iam_role.name}: {eks_iam_role_arn}")
            except Exception as e:
                print_error("IamRole creation failed, please try again")
                print_error(e)
                return False
            if eks_iam_role_arn is None:
                print_error(
                    f"ARN for IamRole {eks_iam_role.name} is not available, please try again"
                )
                return False

            # Create the CloudFormationStack and resources_vpc_config
            vpc_stack = self.get_eks_vpc_stack()
            try:
                vpc_stack.create(aws_client)
                resources_vpc_config = self.get_eks_resources_vpc_config(
                    aws_client, vpc_stack
                )
            except Exception as e:
                print_error("Stack creation failed, please try again")
                print_error(e)
                return False

            # create a dict of args which are not null, otherwise aws type validation fails
            not_null_args = {}

            if self.version:
                not_null_args["version"] = self.version
            if self.kubernetes_network_config:
                not_null_args[
                    "kubernetesNetworkConfig"
                ] = self.kubernetes_network_config
            if self.logging:
                not_null_args["logging"] = self.logging
            if self.client_request_token:
                not_null_args["clientRequestToken"] = self.client_request_token
            if self.tags:
                not_null_args["tags"] = self.tags
            if self.encryption_config:
                not_null_args["encryptionConfig"] = self.encryption_config

            # Create EksCluster
            service_client = self.get_service_client(aws_client)
            create_response = service_client.create_cluster(
                name=self.name,
                roleArn=eks_iam_role_arn,
                resourcesVpcConfig=resources_vpc_config,
                **not_null_args,
            )
            # logger.debug(f"EksCluster: {create_response}")
            # logger.debug(f"EksCluster type: {type(create_response)}")

            # Validate Cluster creation
            self.created_at = create_response.get("cluster", {}).get("createdAt", None)
            self.cluster_status = create_response.get("cluster", {}).get("status", None)
            logger.debug(f"created_at: {self.created_at}")
            logger.debug(f"cluster_status: {self.cluster_status}")
            if self.created_at is not None:
                print_info(f"EksCluster created: {self.name}")
                self.active_resource = create_response
                return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error("Please try again.")
            print_error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:
        # Wait for Cluster to be created
        if self.wait_for_creation:
            try:
                print_info(
                    f"Waiting for {self.get_resource_type()} to be created, this may take a while"
                )
                waiter = self.get_service_client(aws_client).get_waiter(
                    "cluster_active"
                )
                waiter.wait(
                    name=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                print_error("Waiter failed.")
                print_error("Please try again.")
                print_error(e)
        # Update kubeconfig if needed
        if self.update_kubeconfig:
            self.load_kubeconfig(aws_client)
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the EksCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_client = self.get_service_client(aws_client)
            describe_response = service_client.describe_cluster(name=self.name)
            # logger.debug(f"EksCluster: {describe_response}")
            # logger.debug(f"EksCluster type: {type(describe_response)}")

            self.created_at = describe_response.get("cluster", {}).get(
                "createdAt", None
            )
            self.cluster_status = describe_response.get("cluster", {}).get(
                "status", None
            )
            logger.debug(f"EksCluster created_at: {self.created_at}")
            logger.debug(f"EksCluster status: {self.cluster_status}")
            if self.created_at is not None:
                logger.debug(f"EksCluster found: {self.name}")
                self.active_resource = describe_response
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
            pass
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error("Please try again.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the EksCluster
        Deletes the Amazon EKS cluster control plane.
        If you have active services in your cluster that are associated with a load balancer,
        you must delete those services before deleting the cluster so that the load balancers
        are deleted properly. Otherwise, you can have orphaned resources in your VPC
        that prevent you from being able to delete the VPC.

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # Delete the IamRole
            eks_iam_role = self.get_eks_iam_role()
            try:
                print_info(f"Deleting IamRole: {eks_iam_role.name}")
                eks_iam_role.delete(aws_client)
            except Exception as e:
                print_error(
                    "IamRole deletion failed, please try again or delete manually"
                )
                print_error(e)

            # Delete the CloudFormationStack if needed
            resources_vpc_config = self.resources_vpc_config
            if resources_vpc_config is None:
                # indicated vpc stack is created by this cluster
                vpc_stack = self.get_eks_vpc_stack()
                try:
                    print_info(f"Deleting Stack: {vpc_stack.name}")
                    vpc_stack.delete(aws_client)
                except Exception as e:
                    print_error(
                        "Stack deletion failed, please try again or delete manually"
                    )
                    print_error(e)

            # Delete the EksCluster
            service_client = self.get_service_client(aws_client)
            self.active_resource = None

            delete_response = service_client.delete_cluster(name=self.name)
            # logger.debug(f"EksCluster: {delete_cluster_response}")
            # logger.debug(f"EksCluster type: {type(delete_cluster_response)}")
            print_info(
                f"{self.get_resource_type()}: {self.get_resource_name()} deleted"
            )
            return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be deleted.")
            print_error("Please try again or delete resources manually.")
            print_error(e)
        return False

    def post_delete(self, aws_client: AwsApiClient) -> bool:
        # Wait for Cluster to be deleted
        if self.wait_for_deletion:
            try:
                print_info(
                    f"Waiting for {self.get_resource_type()} to be created, this may take a while"
                )
                waiter = self.get_service_client(aws_client).get_waiter(
                    "cluster_deleted"
                )
                waiter.wait(
                    name=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                print_error("Waiter failed.")
                print_error("Please try again.")
                print_error(e)
        # Update kubeconfig if needed
        if self.update_kubeconfig:
            return self.clean_kubeconfig(aws_client)
        return True

    def get_eks_iam_role(self) -> IamRole:
        if self.role is not None:
            return self.role
        return IamRole(
            name=self.role_name or f"{self.name}-role",
            assume_role_policy_document=dedent(
                """\
            {
              "Version": "2012-10-17",
              "Statement": [
                {
                  "Effect": "Allow",
                  "Principal": {
                    "Service": "eks.amazonaws.com"
                  },
                  "Action": "sts:AssumeRole"
                }
              ]
            }
            """
            ),
            policy_arns=["arn:aws:iam::aws:policy/AmazonEKSClusterPolicy"],
        )

    def get_eks_vpc_stack(self) -> CloudFormationStack:
        if self.vpc_stack is not None:
            return self.vpc_stack
        return CloudFormationStack(
            name=self.vpc_stack_name or f"{self.name}-vpc-stack",
            template_url=self.vpc_stack_template_url,
        )

    def get_eks_resources_vpc_config(
        self, aws_client: AwsApiClient, vpc_stack: CloudFormationStack
    ) -> Dict[str, Union[List[str], bool]]:
        if self.resources_vpc_config is not None:
            return self.resources_vpc_config

        # Build resources_vpc_config using vpc_stack
        # get the VPC physical_resource_id
        # vpc_stack_resource = vpc_stack.get_stack_resource(aws_client, "VPC")
        # vpc_physical_resource_id = (
        #     vpc_stack_resource.physical_resource_id
        #     if vpc_stack_resource is not None
        #     else None
        # )
        # # logger.debug(f"vpc_physical_resource_id: {vpc_physical_resource_id}")

        # get the ControlPlaneSecurityGroup physical_resource_id
        sg_stack_resource = vpc_stack.get_stack_resource(
            aws_client, "ControlPlaneSecurityGroup"
        )
        sg_physical_resource_id = (
            sg_stack_resource.physical_resource_id
            if sg_stack_resource is not None
            else None
        )
        # logger.debug(f"sg_physical_resource_id: {sg_physical_resource_id}")

        private_subnets: Optional[List[str]] = vpc_stack.get_private_subnets(aws_client)
        public_subnets: Optional[List[str]] = vpc_stack.get_public_subnets(aws_client)
        subnet_ids: List[str] = []
        if private_subnets is not None:
            subnet_ids.extend(private_subnets)
        if public_subnets is not None:
            subnet_ids.extend(public_subnets)

        resources_vpc_config = {
            "subnetIds": subnet_ids,
            "securityGroupIds": [sg_physical_resource_id],
        }
        return resources_vpc_config

    def get_kubeconfig_cluster_name(self) -> str:
        return self.kubeconfig_cluster_name or f"{self.name}-eks"

    def get_kubeconfig_user_name(self) -> str:
        return self.kubeconfig_cluster_user or f"{self.name}-eks"

    def get_kubeconfig_context_name(self) -> str:
        return self.kubeconfig_cluster_context or f"{self.name}-eks"

    def load_kubeconfig(self, aws_client: AwsApiClient) -> None:

        eks_cluster = self._read(aws_client)
        try:
            cluster_cert = (
                eks_cluster.get("cluster", {})
                .get("certificateAuthority", {})
                .get("data", None)
            )
            # logger.debug(f"cluster_cert: {cluster_cert}")
            cluster_endpoint = eks_cluster.get("cluster", {}).get("endpoint", None)
            # logger.debug(f"cluster_endpoint: {cluster_endpoint}")
            cluster_arn = eks_cluster.get("cluster", {}).get("arn", None)
            # logger.debug(f"cluster_arn: {cluster_arn}")
        except Exception as e:
            print_error(f"Cannot read EKSCluster")
            print_error(e)
            return

        import yaml
        from phidata.infra.k8s.enums.api_version import ApiVersion
        from phidata.infra.k8s.enums.kind import Kind
        from phidata.infra.k8s.resource.kubeconfig import (
            Kubeconfig,
            KubeconfigCluster,
            KubeconfigClusterConfig,
            KubeconfigContext,
            KubeconfigContextSpec,
            KubeconfigUser,
        )

        kubeconfig: Optional[Kubeconfig] = None
        new_cluster = KubeconfigCluster(
            name=self.get_kubeconfig_cluster_name(),
            cluster=KubeconfigClusterConfig(
                server=str(cluster_endpoint),
                certificate_authority_data=str(cluster_cert),
            ),
        )
        new_user = KubeconfigUser(
            name=self.get_kubeconfig_user_name(),
            user={
                "exec": {
                    "apiVersion": ApiVersion.CLIENT_AUTHENTICATION_V1ALPHA1.value,
                    "command": "aws",
                    "args": [
                        "eks",
                        "get-token",
                        "--cluster-name",
                        self.name,
                    ],
                }
            },
        )
        new_context = KubeconfigContext(
            name=self.get_kubeconfig_context_name(),
            context=KubeconfigContextSpec(
                cluster=new_cluster.name,
                user=new_user.name,
            ),
        )
        current_context = new_context.name
        cluster_config: KubeconfigCluster

        # If write_kubeconfig = False, then no changes to kubeconfig needed
        write_kubeconfig = False
        if (
            self.kubeconfig_path is not None
            and self.kubeconfig_path.exists()
            and self.kubeconfig_path.is_file()
        ):
            try:
                print_info(f"Reading kubeconfig at {self.kubeconfig_path}")
                kubeconfig_dict = yaml.safe_load(self.kubeconfig_path.read_text())

                kubeconfig = Kubeconfig(**kubeconfig_dict)

                # Update Kubeconfig.clusters:
                # If a cluster with the same name exists in Kubeconfig.clusters -
                #   check if server and cert values match, if not, remove the existing cluster
                #   and add the new cluster config. Mark cluster_config_exists = True
                # If a cluster with the same name does not exist in Kubeconfig.clusters -
                #   add the new cluster config
                cluster_config_exists = False
                for idx, _cluster in enumerate(kubeconfig.clusters, start=0):
                    if _cluster.name == new_cluster.name:
                        cluster_config_exists = True
                        if (
                            _cluster.cluster.server != new_cluster.cluster.server
                            or _cluster.cluster.certificate_authority_data
                            != new_cluster.cluster.certificate_authority_data
                        ):
                            logger.debug(
                                "Kubeconfig.cluster mismatch, updating cluster config"
                            )
                            removed_cluster_config = kubeconfig.clusters.pop(idx)
                            # logger.debug(
                            #     f"removed_cluster_config: {removed_cluster_config}"
                            # )
                            kubeconfig.clusters.append(new_cluster)
                            write_kubeconfig = True
                if not cluster_config_exists:
                    logger.debug("Adding Kubeconfig.cluster")
                    kubeconfig.clusters.append(new_cluster)
                    write_kubeconfig = True

                # Update Kubeconfig.users:
                # If a user with the same name exists in Kubeconfig.users -
                #   check if user spec matches, if not, remove the existing user
                #   and add the new user config. Mark user_config_exists = True
                # If a user with the same name does not exist in Kubeconfig.users -
                #   add the new user config
                user_config_exists = False
                for idx, _user in enumerate(kubeconfig.users, start=0):
                    if _user.name == new_user.name:
                        user_config_exists = True
                        if _user.user != new_user.user:
                            logger.debug(
                                "Kubeconfig.user mismatch, updating user config"
                            )
                            removed_user_config = kubeconfig.users.pop(idx)
                            # logger.debug(f"removed_user_config: {removed_user_config}")
                            kubeconfig.users.append(new_user)
                            write_kubeconfig = True
                if not user_config_exists:
                    logger.debug("Adding Kubeconfig.user")
                    kubeconfig.users.append(new_user)
                    write_kubeconfig = True

                # Update Kubeconfig.contexts:
                # If a context with the same name exists in Kubeconfig.contexts -
                #   check if context spec matches, if not, remove the existing context
                #   and add the new context. Mark context_config_exists = True
                # If a context with the same name does not exist in Kubeconfig.contexts -
                #   add the new context config
                context_config_exists = False
                for idx, _context in enumerate(kubeconfig.contexts, start=0):
                    if _context.name == new_context.name:
                        context_config_exists = True
                        if _context.context != new_context.context:
                            logger.debug(
                                "Kubeconfig.context mismatch, updating context config"
                            )
                            removed_context_config = kubeconfig.contexts.pop(idx)
                            # logger.debug(
                            #     f"removed_context_config: {removed_context_config}"
                            # )
                            kubeconfig.contexts.append(new_context)
                            write_kubeconfig = True
                if not context_config_exists:
                    logger.debug("Adding Kubeconfig.context")
                    kubeconfig.contexts.append(new_context)
                    write_kubeconfig = True

                if (
                    kubeconfig.current_context is None
                    or kubeconfig.current_context != current_context
                ):
                    logger.debug("Updating Kubeconfig.current_context")
                    kubeconfig.current_context = current_context
                    write_kubeconfig = True
            except Exception as e:
                print_error(f"Cannot update kubeconfig at {self.kubeconfig_path}")
                print_error(e)
                kubeconfig = None

        # Create new kubeconfig
        if kubeconfig is None:
            print_info(f"Creating new kubeconfig")
            kubeconfig = Kubeconfig(
                api_version=ApiVersion.CORE_V1,
                kind=Kind.CONFIG,
                clusters=[new_cluster],
                users=[new_user],
                contexts=[new_context],
                current_context=current_context,
            )
            write_kubeconfig = True

        # if kubeconfig:
        #     logger.debug("Kubeconfig:\n{}".format(kubeconfig.json(exclude_none=True, by_alias=True, indent=4)))
        if write_kubeconfig:
            try:
                print_info("Writing kubeconfig")
                kubeconfig_dict = kubeconfig.dict(exclude_none=True, by_alias=True)
                self.kubeconfig_path.write_text(yaml.safe_dump(kubeconfig_dict))
                print_info("Kubeconfig updated")
            except Exception as e:
                print_error(f"Cannot write to {self.kubeconfig_path}")
                print_error(e)
        else:
            print_info("Kubeconfig up-to-date")

    def clean_kubeconfig(self, aws_client: AwsApiClient) -> bool:
        logger.debug(f"TO_DO: Cleaning kubeconfig at {str(self.kubeconfig_path)}")
        return True



from typing import Optional, Any, Union, List, Dict, Tuple
from typing_extensions import Literal

from sqlalchemy.engine import Engine, Connection

from phidata.asset import DataAsset, DataAssetArgs
from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger


class SqlType(ExtendedEnum):
    POSTGRES = "POSTGRES"


class SqlTableArgs(DataAssetArgs):
    # Table Name
    name: str
    # Table schema
    db_schema: Optional[str] = None
    # Type of SQL table
    sql_type: Optional[SqlType] = None
    # airflow connection_id can be used for running workflows remotely
    db_conn_id: Optional[str] = None
    # a db_conn_url can be used to create the sqlalchemy.engine.Engine object
    db_conn_url: Optional[str] = None
    # sqlalchemy.engine.(Engine or Connection)
    # Using SQLAlchemy makes it possible to use any DB supported by that library.
    # NOTE: db_engine is required but can be derived using
    # db_conn_url when running locally and
    # db_conn_id when running remotely
    db_engine: Optional[Union[Engine, Connection]] = None


class SqlTable(DataAsset):
    """Base Class for Sql tables"""

    def __init__(self) -> None:
        super().__init__()
        self.args: Optional[SqlTableArgs] = None

    @property
    def db_schema(self) -> Optional[str]:
        return self.args.db_schema if self.args else None

    @db_schema.setter
    def db_schema(self, db_schema: Optional[str]) -> None:
        if db_schema is not None:
            self.args.db_schema = db_schema

    @property
    def sql_type(self) -> Optional[SqlType]:
        return self.args.sql_type if self.args else None

    @property
    def db_conn_id(self) -> Optional[str]:
        return self.args.db_conn_id if self.args else None

    @db_conn_id.setter
    def db_conn_id(self, db_conn_id: str) -> None:
        if db_conn_id is not None:
            self.args.db_conn_id = db_conn_id

    @property
    def db_conn_url(self) -> Optional[str]:
        return self.args.db_conn_url if self.args else None

    @db_conn_url.setter
    def db_conn_url(self, db_conn_url: str) -> None:
        if db_conn_url is not None:
            self.args.db_conn_url = db_conn_url

    @property
    def db_engine(self) -> Optional[Union[Engine, Connection]]:
        return self.args.db_engine if self.args else None

    @db_engine.setter
    def db_engine(self, db_engine: Optional[Union[Engine, Connection]]) -> None:
        if db_engine is not None:
            self.args.db_engine = db_engine

    def create_db_engine_using_conn_url(self) -> None:
        # Create the SQLAlchemy engine using db_conn_url

        if self.db_conn_url is None:
            return

        try:
            from sqlalchemy import create_engine

            logger.info(f"Creating db_engine using db_conn_url: {self.db_conn_url}")
            db_engine = (create_engine(self.db_conn_url),)
            if isinstance(db_engine, tuple) and len(db_engine) > 0:
                self.db_engine = db_engine[0]
            else:
                self.db_engine = db_engine
        except Exception as e:
            logger.error(f"Error creating db_engine using {self.db_conn_url}")
            logger.error(e)
            return

    def create_db_engine_using_conn_id(self) -> None:
        # Create the SQLAlchemy engine using db_conn_id

        if self.db_conn_id is None:
            return

        try:
            from airflow.providers.postgres.hooks.postgres import PostgresHook

            logger.info(f"Creating db_engine using db_conn_id: {self.db_conn_id}")
            if self.sql_type == SqlType.POSTGRES:
                pg_hook = PostgresHook(postgres_conn_id=self.db_conn_id)
                self.db_engine = pg_hook.get_sqlalchemy_engine()
        except Exception as e:
            logger.error(f"Error creating db_engine using {self.db_conn_id}")
            logger.error(e)
            return

    ######################################################
    ## Write table
    ######################################################

    def write_pandas_df(
        self,
        df: Optional[Any] = None,
        # How to behave if the table already exists.
        # fail: Raise a ValueError.
        # replace: Drop the table before inserting new values.
        # append: Insert new values to the existing table.
        if_exists: Optional[Literal["fail", "replace", "append"]] = None,
        # Write DataFrame index as a column.
        # Uses index_label as the column name in the table.
        index: Optional[bool] = None,
        # Column label for index column(s).
        # If None is given (default) and index is True, then the index names are used.
        # A sequence should be given if the DataFrame uses MultiIndex.
        index_label: Optional[Union[str, List[str]]] = None,
        # Specify the number of rows in each batch to be written at a time.
        # By default, all rows will be written at once.
        chunksize: Optional[int] = None,
    ) -> bool:
        """
        Write DataFrame to table.

        Args:
            index_col : str or list of str, optional, default: None
                Column(s) to set as index(MultiIndex).
            coerce_float : bool, default True
                Attempts to convert values of non-string, non-numeric objects (like
                decimal.Decimal) to floating point. Can result in loss of Precision.
            parse_dates : list or dict, default None
                - List of column names to parse as dates.
                - Dict of ``{column_name: format string}`` where format string is
                strftime compatible in case of parsing string times or is one of
                (D, s, ns, ms, us) in case of parsing integer timestamps.
                - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
                to the keyword arguments of :func:`pandas.to_datetime`
                Especially useful with databases without native Datetime support,
                such as SQLite.
            columns : list, default None
                List of column names to select from SQL table.
            chunksize : int, default None
                If specified, returns an iterator where `chunksize` is the number of
                rows to include in each chunk.

        Returns:
            DataFrame or Iterator[DataFrame]
            A SQL table is returned as two-dimensional data structure with labeled
            axes.
        """

        # SqlTable not yet initialized
        if self.args is None:
            return False

        # Check name is available
        if self.name is None:
            logger.error("SqlTable name not available")
            return False

        # Check engine is available
        if self.db_engine is None:
            if self.db_conn_url is not None:
                self.create_db_engine_using_conn_url()
            elif self.db_conn_id is not None:
                self.create_db_engine_using_conn_id()
            if self.db_engine is None:
                logger.error("DbEngine not available")
                return False

        # write to table
        import pandas as pd

        if df is None or not isinstance(df, pd.DataFrame):
            logger.error("DataFrame invalid")
            return False

        logger.info("Writing to table: {}".format(self.name))

        # create a dict of args provided
        non_null_args = {}
        if self.db_schema:
            non_null_args["schema"] = self.db_schema
        if if_exists:
            non_null_args["if_exists"] = if_exists
        if index:
            non_null_args["index"] = index
        if index_label:
            non_null_args["index_label"] = index_label
        if chunksize:
            non_null_args["chunksize"] = chunksize

        try:
            with self.db_engine.connect() as connection:
                df.to_sql(
                    name=self.name,
                    con=connection,
                    **non_null_args,
                )
                logger.info(f"Done")
            return True
        except Exception:
            logger.error("Could not write table: {}".format(self.name))
            raise

    ######################################################
    ## Read table
    ######################################################

    def read_pandas_df(
        self,
        index_col: Optional[Union[str, List[str]]] = None,
        coerce_float: bool = True,
        parse_dates: Optional[Union[List, Dict]] = None,
        columns: Optional[List[str]] = None,
        chunksize: Optional[int] = None,
    ) -> Optional[Any]:
        """
        Read table into a DataFrame.

        Args:
            index_col : str or list of str, optional, default: None
                Column(s) to set as index(MultiIndex).
            coerce_float : bool, default True
                Attempts to convert values of non-string, non-numeric objects (like
                decimal.Decimal) to floating point. Can result in loss of Precision.
            parse_dates : list or dict, default None
                - List of column names to parse as dates.
                - Dict of ``{column_name: format string}`` where format string is
                strftime compatible in case of parsing string times or is one of
                (D, s, ns, ms, us) in case of parsing integer timestamps.
                - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
                to the keyword arguments of :func:`pandas.to_datetime`
                Especially useful with databases without native Datetime support,
                such as SQLite.
            columns : list, default None
                List of column names to select from SQL table.
            chunksize : int, default None
                If specified, returns an iterator where `chunksize` is the number of
                rows to include in each chunk.

        Returns:
            DataFrame or Iterator[DataFrame]
            A SQL table is returned as two-dimensional data structure with labeled
            axes.
        """

        # SqlTable not yet initialized
        if self.args is None:
            return False

        # Check db_engine is available
        if self.db_engine is None:
            if self.db_conn_url is not None:
                self.create_db_engine_using_conn_url()
            elif self.db_conn_id is not None:
                self.create_db_engine_using_conn_id()
            if self.db_engine is None:
                logger.error("DbEngine not available")
                return False

        # read sql table
        import pandas as pd

        logger.info("Reading table: {}".format(self.name))

        # create a dict of args provided
        non_null_args = {}
        if self.db_schema:
            non_null_args["schema"] = self.db_schema
        if index_col:
            non_null_args["index_col"] = index_col
        if coerce_float:
            non_null_args["coerce_float"] = coerce_float
        if parse_dates:
            non_null_args["parse_dates"] = parse_dates
        if columns:
            non_null_args["columns"] = columns
        if chunksize:
            non_null_args["chunksize"] = chunksize

        try:
            with self.db_engine.connect() as connection:
                result_df = pd.read_sql_table(
                    table_name=self.name,
                    con=connection,
                    **non_null_args,
                )
            return result_df
        except Exception:
            logger.error("Could not read table: {}".format(self.name))
            raise

    def run_sql_query(
        self,
        sql_query: str,
        index_col: Optional[Union[str, List[str]]] = None,
        coerce_float: bool = True,
        params: Optional[Union[List, Tuple, Dict]] = None,
        parse_dates: Optional[Union[List, Dict]] = None,
        chunksize: Optional[int] = None,
        dtype: Optional[Any] = None,
    ) -> Optional[Any]:
        """
        Read SQL query into a DataFrame.

        Args:
            index_col : str or list of str, optional, default: None
                Column(s) to set as index(MultiIndex).
            coerce_float : bool, default True
                Attempts to convert values of non-string, non-numeric objects (like
                decimal.Decimal) to floating point. Can result in loss of Precision.
            params : list, tuple or dict, optional, default: None
                List of parameters to pass to execute method.  The syntax used
                to pass parameters is database driver dependent. Check your
                database driver documentation for which of the five syntax styles,
                described in PEP 249's paramstyle, is supported.
                Eg. for psycopg2, uses %(name)s so use params={'name' : 'value'}.
            parse_dates : list or dict, default None
                - List of column names to parse as dates.
                - Dict of ``{column_name: format string}`` where format string is
                strftime compatible in case of parsing string times or is one of
                (D, s, ns, ms, us) in case of parsing integer timestamps.
                - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
                to the keyword arguments of :func:`pandas.to_datetime`
                Especially useful with databases without native Datetime support,
                such as SQLite.
            chunksize : int, default None
                If specified, returns an iterator where `chunksize` is the number of
                rows to include in each chunk.
            dtype : Type name or dict of columns
                Data type for data or columns. E.g. np.float64 or
                {‘a’: np.float64, ‘b’: np.int32, ‘c’: ‘Int64’}

        Returns:
            DataFrame or Iterator[DataFrame]
            A SQL table is returned as two-dimensional data structure with labeled
            axes.
        """

        # SqlTable not yet initialized
        if self.args is None:
            return False

        # Check db_engine is available
        if self.db_engine is None:
            if self.db_conn_url is not None:
                self.create_db_engine_using_conn_url()
            elif self.db_conn_id is not None:
                self.create_db_engine_using_conn_id()
            if self.db_engine is None:
                logger.error("DbEngine not available")
                return False

        # run sql query
        import pandas as pd

        logger.info("Running Query:\n{}".format(sql_query))

        # create a dict of args provided
        non_null_args = {}
        if self.db_schema:
            non_null_args["schema"] = self.db_schema
        if index_col:
            non_null_args["index_col"] = index_col
        if coerce_float:
            non_null_args["coerce_float"] = coerce_float
        if params:
            non_null_args["params"] = params
        if parse_dates:
            non_null_args["parse_dates"] = parse_dates
        if chunksize:
            non_null_args["chunksize"] = chunksize
        if dtype:
            non_null_args["dtype"] = dtype

        try:
            with self.db_engine.connect() as connection:
                result_df = pd.read_sql_query(
                    sql=sql_query,
                    con=connection,
                    **non_null_args,
                )
            return result_df
        except Exception:
            logger.error("Could not run query on: {}".format(self.name))
            raise



from pathlib import Path
from typing import Optional, List

from phidata.infra.docker.args import DockerArgs
from phidata.infra.docker.enums import DockerManagerStatus
from phidata.infra.docker.exceptions import DockerArgsException
from phidata.infra.docker.resource.types import DockerResourceType
from phidata.infra.docker.worker import DockerWorker
from phidata.utils.cli_console import print_error
from phidata.utils.log import logger


class DockerManager:
    def __init__(self, docker_args: DockerArgs):
        if docker_args is None or not isinstance(docker_args, DockerArgs):
            raise DockerArgsException("Invalid DockerArgs")
        if docker_args.workspace_root_path is None or not isinstance(
            docker_args.workspace_root_path, Path
        ):
            raise DockerArgsException("workspace_root_path invalid")
        if docker_args.workspace_config_file_path is None or not isinstance(
            docker_args.workspace_config_file_path, Path
        ):
            raise DockerArgsException("workspace_config_file_path invalid")

        self.docker_args: DockerArgs = docker_args
        self.docker_worker: DockerWorker = DockerWorker(self.docker_args)
        self.docker_status: DockerManagerStatus = DockerManagerStatus.PRE_INIT
        logger.debug("**-+-** DockerManager created")

    def get_status(self, refresh: bool = False) -> DockerManagerStatus:
        # logger.debug("Getting DockerManagerStatus")
        if refresh:
            self.docker_status = DockerManagerStatus.PRE_INIT
            logger.debug(f"DockerManagerStatus: {self.docker_status.value}")

        if self.docker_status == DockerManagerStatus.PRE_INIT:
            if (
                self.docker_worker is not None
                and self.docker_worker.is_client_initialized()
            ):
                self.docker_status = DockerManagerStatus.WORKER_READY
            else:
                print_error("DockerWorker not ready")

        if self.docker_status.can_create_resources():
            if (
                self.docker_worker is not None
                and self.docker_worker.are_resources_initialized()
            ):
                self.docker_status = DockerManagerStatus.RESOURCES_INIT

        logger.debug(f"DockerManagerStatus: {self.docker_status.value}")
        return self.docker_status

    def get_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> Optional[List[DockerResourceType]]:

        status = self.get_status()
        if not status.can_get_resources():
            logger.debug("Cannot get resources")
            return None
        if self.docker_worker is None:
            logger.debug("No worker available")
            return None

        return self.docker_worker.get_resources(name_filter, type_filter)

    def create_resources_dry_run(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> None:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot create resources")
            return
        if self.docker_worker is None:
            logger.debug("No worker available")
            return

        # logger.debug("Creating resources dry run")
        self.docker_worker.create_resources_dry_run(
            name_filter=name_filter, type_filter=type_filter
        )

    def create_resources(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot create resources")
            return False
        if self.docker_worker is None:
            logger.debug("No worker available")
            return False

        return self.docker_worker.create_resources(
            name_filter=name_filter, type_filter=type_filter
        )

    def validate_resources_are_created(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        logger.debug("Validating resources are created...")
        return True

    def delete_resources_dry_run(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> None:

        status = self.get_status()
        if not status.can_delete_resources():
            logger.debug("Cannot delete resources")
            return
        if self.docker_worker is None:
            logger.debug("No worker available")
            return

        # logger.debug("Deleting resources dry run")
        self.docker_worker.delete_resources_dry_run(
            name_filter=name_filter, type_filter=type_filter
        )

    def delete_resources(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        status = self.get_status()
        if not status.can_delete_resources():
            logger.debug("Cannot delete resources")
            return False
        if self.docker_worker is None:
            logger.debug("No worker available")
            return False

        return self.docker_worker.delete_resources(name_filter=name_filter)

    def validate_resources_are_deleted(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        logger.debug("Validating resources are deleted...")
        return True

    ######################################################
    ## Patch Resources
    ######################################################

    def patch_resources_dry_run(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> None:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot patch resources")
            return
        if self.docker_worker is None:
            logger.debug("No worker available")
            return

        # logger.debug("Deleting resources dry run")
        self.docker_worker.patch_resources_dry_run(
            name_filter=name_filter, type_filter=type_filter
        )

    def patch_resources(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot patch resources")
            return False
        if self.docker_worker is None:
            logger.debug("No worker available")
            return False

        return self.docker_worker.patch_resources(name_filter=name_filter)

    def validate_resources_are_patched(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        logger.debug("Validating resources are patched...")
        return True



from pathlib import Path
from typing import Optional, List, Dict

from phidata.infra.k8s.args import K8sArgs
from phidata.infra.k8s.enums import K8sManagerStatus
from phidata.infra.k8s.exceptions import K8sArgsException
from phidata.infra.k8s.resource.types import K8sResourceType
from phidata.infra.k8s.resource.group import K8sResourceGroup
from phidata.infra.k8s.worker import K8sWorker
from phidata.utils.cli_console import print_error, print_info
from phidata.utils.log import logger


class K8sManager:
    def __init__(self, k8s_args: K8sArgs):
        if k8s_args is None or not isinstance(k8s_args, K8sArgs):
            raise K8sArgsException("Invalid K8sArgs")
        if k8s_args.workspace_root_path is None or not isinstance(
            k8s_args.workspace_root_path, Path
        ):
            raise K8sArgsException("workspace_root_path invalid")
        if k8s_args.workspace_config_file_path is None or not isinstance(
            k8s_args.workspace_config_file_path, Path
        ):
            raise K8sArgsException("workspace_config_file_path invalid")

        self.k8s_args: K8sArgs = k8s_args
        self.k8s_worker: Optional[K8sWorker] = K8sWorker(self.k8s_args)
        self.k8s_status: K8sManagerStatus = K8sManagerStatus.PRE_INIT
        logger.debug("**-+-** K8sManager created")

    def get_status(self, refresh: bool = False) -> K8sManagerStatus:
        # logger.debug("Getting K8sManagerStatus")
        if refresh:
            self.k8s_status = K8sManagerStatus.PRE_INIT
            logger.debug(f"K8sManagerStatus: {self.k8s_status.value}")

        if self.k8s_status == K8sManagerStatus.PRE_INIT:
            if self.k8s_worker is not None and self.k8s_worker.is_client_initialized():
                self.k8s_status = K8sManagerStatus.WORKER_READY

        if self.k8s_status == K8sManagerStatus.WORKER_READY:
            if (
                self.k8s_worker is not None
                and self.k8s_worker.are_resources_initialized()
            ):
                self.k8s_status = K8sManagerStatus.RESOURCES_INIT

        logger.debug(f"K8sManagerStatus: {self.k8s_status.value}")
        return self.k8s_status

    ######################################################
    ## Create Resources
    ######################################################

    def create_resources_dry_run(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> None:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot create resources")
            return
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return

        # logger.debug("Creating resources dry run")
        self.k8s_worker.create_resources_dry_run(
            name_filter=name_filter, type_filter=type_filter
        )

    def create_resources(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot create resources")
            return False
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return False

        return self.k8s_worker.create_resources(
            name_filter=name_filter, type_filter=type_filter
        )

    def validate_resources_are_created(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        logger.debug("Validating resources are created...")
        return True

    ######################################################
    ## Read Resources
    ######################################################

    def get_resources(self) -> Optional[Dict[str, K8sResourceGroup]]:

        status = self.get_status()
        if not status.can_get_resources():
            logger.debug("Cannot get resources")
            return None
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return None

        return self.k8s_worker.get_resources()

    def read_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> Optional[List[K8sResourceType]]:

        status = self.get_status()
        if not status.can_read_resources():
            logger.debug("Cannot get resources")
            return None
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return None

        return self.k8s_worker.read_resources(name_filter, type_filter)

    ######################################################
    ## Delete Resources
    ######################################################

    def delete_resources_dry_run(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> None:

        status = self.get_status()
        if not status.can_delete_resources():
            logger.debug("Cannot delete resources")
            return
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return

        # logger.debug("Deleting resources dry run")
        self.k8s_worker.delete_resources_dry_run(
            name_filter=name_filter, type_filter=type_filter
        )

    def delete_resources(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        status = self.get_status()
        if not status.can_delete_resources():
            logger.debug("Cannot delete resources")
            return False
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return False

        return self.k8s_worker.delete_resources(
            name_filter=name_filter, type_filter=type_filter
        )

    def validate_resources_are_deleted(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        logger.debug("Validating resources are deleted...")
        return True

    ######################################################
    ## Patch Resources
    ######################################################

    def patch_resources_dry_run(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> None:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot patch resources")
            return
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return

        # logger.debug("Deleting resources dry run")
        self.k8s_worker.patch_resources_dry_run(
            name_filter=name_filter, type_filter=type_filter
        )

    def patch_resources(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot patch resources")
            return False
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return False

        return self.k8s_worker.patch_resources(
            name_filter=name_filter, type_filter=type_filter
        )

    def validate_resources_are_patched(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        logger.debug("Validating resources are patched...")
        return True



import datetime
from typing import List, Optional, Any, Dict

from pydantic import BaseModel, Field

import phidata.infra.k8s.enums as k8s_enums


# Kubeconfig Cluster
class KubeconfigClusterConfig(BaseModel):
    server: str
    certificate_authority_data: str = Field(..., alias="certificate-authority-data")

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        use_enum_values = True


class KubeconfigCluster(BaseModel):
    name: str
    cluster: KubeconfigClusterConfig


class KubeconfigUser(BaseModel):
    name: str
    user: Dict[str, Any]


# Kubeconfig Context
class KubeconfigContextSpec(BaseModel):
    """Each Kubeconfig context is a triple (cluster, user, namespace). It should be read as:
    Use the credentials of the "user" to access the "namespace" of the "cluster”
    """

    cluster: Optional[str]
    user: Optional[str]
    namespace: Optional[str]


class KubeconfigContext(BaseModel):
    """A context element in a kubeconfig file is used to group access parameters under a
    convenient name. Each context has three parameters: cluster, namespace, and user.
    By default, the kubectl command-line tool uses parameters from the current context
    to communicate with the cluster.
    """

    name: str
    context: KubeconfigContextSpec


# Kubeconfig
class Kubeconfig(BaseModel):
    """This is one of the most important models, and the one we know the least about :(

    We configure access to K8s clusters using a Kubeconfig. This configuration can be stored in a file or an object.
    A Kubeconfig stores information about clusters, users, namespaces, and authentication mechanisms,

    When we use K8s locally on our machine, our kubeconfig file is usually stored at ~/.kube/config
    A file that is used to configure access to a cluster is called a kubeconfig file. This is a generic way of referring to configuration files. It does not mean that there is a file named kubeconfig.
    View your local kubeconfig using `kubectl config view`

    References:
        * Docs:
            https://kubernetes.io/docs/tasks/access-application-cluster/configure-access-multiple-clusters/
        * Go Doc: https://godoc.org/k8s.io/client-go/tools/clientcmd/api#Config
    """

    api_version: k8s_enums.ApiVersion = Field(..., alias="apiVersion")
    kind: k8s_enums.Kind
    clusters: List[KubeconfigCluster] = []
    users: List[KubeconfigUser] = []
    contexts: List[KubeconfigContext] = []
    current_context: Optional[str] = Field(None, alias="current-context")
    preferences: dict = {}

    class Config:
        allow_population_by_field_name = True
        use_enum_values = True



from typing import Optional, Callable, List, Any, Dict

from phidata.utils.cli_console import print_error
from phidata.workflow.py.python_workflow import (
    PythonWorkflow,
    PythonWorkflowArgs,
    EngineType,
)


class PythonFunction(PythonWorkflow):
    def __init__(
        self,
        name: str,
        entrypoint: Callable[..., bool],
        entrypoint_args: Optional[List] = None,
        entrypoint_kwargs: Optional[Dict] = None,
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
        engine: EngineType = EngineType.DEFAULT,
        # If True, scans the args for SqlTable objects and populates the
        # SqlTable.db_engine value using the SqlTable.db_conn_id
        create_db_engine_from_conn_id: bool = False,
    ) -> None:
        super().__init__()
        try:
            self.args = PythonWorkflowArgs(
                name=name,
                entrypoint=entrypoint,
                entrypoint_args=entrypoint_args,
                entrypoint_kwargs=entrypoint_kwargs,
                task_id=task_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                engine=engine,
                create_db_engine_from_conn_id=create_db_engine_from_conn_id,
            )
        except Exception as e:
            print_error(f"Args for {self.__class__.__name__} are not valid")
            raise



from typing import Optional, List, Dict, Tuple

from typing_extensions import Literal

from phidata.infra.docker.resource.base import DockerResource
from phidata.infra.docker.resource.group import DockerResourceGroup
from phidata.infra.docker.resource.types import (
    DockerResourceInstallOrder,
    DockerResourceType,
)
from phidata.utils.log import logger


def get_install_weight_for_docker_resource(
    docker_resource_type: DockerResourceType, resource_group_weight: int = 100
) -> int:
    """Function which takes a DockerResource and resource_group_weight and returns the install
    weight for that resource.

    Understanding install weights for DockerResources:

    - Each DockerResource gets an install weight, which determines the order for installing that particular resource.
    - By default, DockerResources are installed in the order determined by the DockerResourceInstallOrder dict.
        The DockerResourceInstallOrder dict ensures volumes are created before containers
    - We can also provide a weight to the DockerResourceGroup, that weight determines the install order for resources of
        that resource group as compared to other resource groups.
    - We achieve this by multiplying the DockerResourceGroup.weight with the value from DockerResourceInstallOrder
        and by default using a weight of 100. So
        * Weights 1-10 are reserved
        * Choose weight 11-99 to deploy a resource group before all the "default" resources
        * Choose weight 101+ to deploy a resource group after all the "default" resources
        * Choosing weight 100 has no effect because that is the default install weight
    """
    resource_type_class_name = docker_resource_type.__class__.__name__
    if resource_type_class_name in DockerResourceInstallOrder.keys():
        install_weight = DockerResourceInstallOrder[resource_type_class_name]
        final_weight = resource_group_weight * install_weight
        # logger.debug(
        #     "Resource type: {} | RG Weight: {} | Install weight: {} | Final weight: {}".format(
        #         resource_type_class_name,
        #         resource_group_weight,
        #         install_weight,
        #         final_weight,
        #     )
        # )
        return final_weight

    return 5000


def get_docker_resources_from_group(
    docker_resource_group: DockerResourceGroup,
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
) -> Optional[List[DockerResourceType]]:
    """Parses the DockerResourceGroup and returns an array of DockerResources
    after applying the name & type filters. This function also flattens any
    List[DockerResourceType] attributes. Eg: it will flatten List[DockerContainer]

    Args:
        docker_resource_group:
        name_filter: filter DockerResources by name
        type_filter: filter DockerResources by type

    Returns:
        List[DockerResourceType]: List of filtered and flattened DockerResources
    """

    # List of resources to return, we use the Union type: DockerResourceType
    # to hold all DockerResources
    docker_resources: List[DockerResourceType] = []
    # logger.debug(f"Flattening {docker_resource_group.name}")

    # populate the docker_resources list with the resources
    # Loop over each key of the DockerResourceGroup model
    for resource_key, resource_data in docker_resource_group.__dict__.items():
        # logger.debug("resource: {}".format(resource))
        # logger.debug("resource_data: {}".format(resource_data))

        # Check if the resource is a single DockerResource or a List[DockerResource]
        # If it is a List[DockerResource], flatten the resources, verify each element
        # of the list is a subclass of DockerResource and add to the docker_resources list
        if isinstance(resource_data, list):
            for _r in resource_data:
                if isinstance(_r, DockerResource):
                    if name_filter is not None:
                        # logger.debug(f"name_filter: {name_filter.lower()}")
                        # logger.debug(f"resource name: {_r.get_resource_name().lower()}")
                        if name_filter.lower() not in _r.get_resource_name().lower():
                            # logger.debug(
                            #     f"skipping {_r.get_resource_type()}:{_r.get_resource_name()}"
                            # )
                            continue

                    if type_filter is not None:
                        # logger.debug(f"type_filter: {type_filter.lower()}")
                        # logger.debug(f"class: {_r.get_resource_type().lower()}")
                        if type_filter.lower() not in _r.get_resource_type().lower():
                            # logger.debug(
                            #     f"skipping {_r.get_resource_type()}:{_r.get_resource_name()}"
                            # )
                            continue
                    docker_resources.append(_r)  # type: ignore

        # If its a single resource, verify that the resource is a subclass of
        # DockerResource and add it to the docker_resources list
        elif isinstance(resource_data, DockerResource):
            if name_filter is not None:
                if name_filter.lower() not in resource_data.get_resource_name().lower():
                    # logger.debug(
                    #     f"skipping {resource_data.get_resource_type()}:{resource_data.get_resource_name()}"
                    # )
                    continue

            if type_filter is not None:
                # logger.debug(f"type_filter: {type_filter.lower()}")
                # logger.debug(f"class: {resource_data.get_resource_type().lower()}")
                if type_filter.lower() not in resource_data.get_resource_type().lower():
                    # logger.debug(
                    #     f"skipping {resource_data.get_resource_type()}:{resource_data.get_resource_name()}"
                    # )
                    continue
            docker_resources.append(resource_data)  # type: ignore

    return docker_resources


def dedup_docker_resources(
    docker_resources_with_weight: List[Tuple[DockerResourceType, int]],
) -> List[Tuple[DockerResourceType, int]]:
    if docker_resources_with_weight is None:
        raise ValueError

    deduped_resources: List[Tuple[DockerResourceType, int]] = []
    prev_rsrc: Optional[DockerResourceType] = None
    prev_weight: Optional[int] = None
    for rsrc, weight in docker_resources_with_weight:
        # First item of loop
        if prev_rsrc is None:
            prev_rsrc = rsrc
            prev_weight = weight
            deduped_resources.append((rsrc, weight))
            continue

        # Compare resources with same weight only
        if weight == prev_weight:
            if (
                rsrc.get_resource_type() == prev_rsrc.get_resource_type()
                and rsrc.get_resource_name() == prev_rsrc.get_resource_name()
                and rsrc.get_resource_name() is not None
            ):
                # If resource type and name are the same, skip the resource
                # Note: resource.name cannot be None
                continue

        # If the loop hasn't been continued by the if blocks above,
        # add the resource and weight to the deduped_resources list
        deduped_resources.append((rsrc, weight))
        # update the previous resource for comparison
        prev_rsrc = rsrc
        prev_weight = weight
    return deduped_resources


def filter_and_flatten_docker_resource_groups(
    docker_resource_groups: Dict[str, DockerResourceGroup],
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
    sort_order: Literal["create", "delete"] = "create",
) -> Optional[List[DockerResourceType]]:
    """This function parses the docker_resource_groups dict and returns a filtered array of
    DockerResources sorted in the order requested. create == install order, delete == reverse
    Desc:
        1. Iterate through each DockerResourceGroup
        2. If group is enabled, get the DockerResources from that group
        3. Return a list of all DockerResources from 2.

    Args:
        docker_resource_groups: Dict[str, DockerResourceGroup]
            Dict of {resource_group_name : DockerResourceGroup}
        name_filter: Filter DockerResourceGroup by name
        type_filter: Filter DockerResourceGroup by type
        sort_order: create or delete

    Returns:
        List[DockerResourceType]: List of filtered DockerResources
    """

    # The list of DockerResources that will be returned
    filtered_docker_resources: List[DockerResourceType] = []
    logger.debug("Flattening DockerResourceGroups")

    # Step 1: Create docker_resource_list_with_weight
    # A List of Tuples where each tuple is a (DockerResource, Resource Group Weight)
    # This list helps us sort the DockerResources
    # based on their resource group weight using get_install_weight_for_docker_resource
    docker_resource_list_with_weight: List[Tuple[DockerResourceType, int]] = []
    if docker_resource_groups:
        # Iterate through docker_resource_groups
        for docker_rg_name, docker_rg in docker_resource_groups.items():
            # logger.debug("docker_rg_name: {}".format(docker_rg_name))
            # logger.debug("docker_rg: {}".format(docker_rg))

            # skip disabled DockerResourceGroups
            if not docker_rg.enabled:
                continue

            docker_resources = get_docker_resources_from_group(
                docker_rg, name_filter, type_filter
            )
            if docker_resources:
                for _docker_rsrc in docker_resources:
                    docker_resource_list_with_weight.append(
                        (_docker_rsrc, docker_rg.weight)
                    )

    # Sort the resources in install order
    if sort_order == "create":
        docker_resource_list_with_weight.sort(
            key=lambda x: get_install_weight_for_docker_resource(x[0], x[1])
        )
    elif sort_order == "delete":
        docker_resource_list_with_weight.sort(
            key=lambda x: get_install_weight_for_docker_resource(x[0], x[1]),
            reverse=True,
        )

    # De-duplicate DockerResources
    # Mainly used to remove the extra Network resources
    deduped_docker_resources: List[
        Tuple[DockerResourceType, int]
    ] = dedup_docker_resources(docker_resource_list_with_weight)
    # deduped_docker_resources = docker_resource_list_with_weight
    # logger.debug("DockerResource list")
    # logger.debug(
    #     "\n".join(
    #         "Name: {}, Type: {}, Weight: {}".format(
    #             rsrc.name, rsrc.resource_type, weight
    #         )
    #         for rsrc, weight in deduped_docker_resources
    #     )
    # )

    # drop the weight from the deduped_docker_resources tuple
    filtered_docker_resources = [x[0] for x in deduped_docker_resources]
    # logger.debug("filtered_docker_resources: {}".format(filtered_docker_resources))
    return filtered_docker_resources



from typing import Optional

from phidata.asset.aws.glue.crawler import GlueCrawler
from phidata.utils.log import logger
from phidata.workflow.py.python_workflow import (
    PythonWorkflow,
    PythonWorkflowArgs,
)


class CreateGlueCrawlerArgs(PythonWorkflowArgs):
    crawler: GlueCrawler
    start_crawler: bool = False


class CreateGlueCrawler(PythonWorkflow):
    def __init__(
        self,
        crawler: GlueCrawler,
        start_crawler: bool = False,
        name: str = "create_glue_crawler",
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ):
        super().__init__()
        try:
            self.args: CreateGlueCrawlerArgs = CreateGlueCrawlerArgs(
                crawler=crawler,
                start_crawler=start_crawler,
                name=name,
                task_id=task_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                entrypoint=create_glue_crawler,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def crawler(self) -> GlueCrawler:
        return self.args.crawler

    @crawler.setter
    def crawler(self, crawler: GlueCrawler) -> None:
        if crawler is not None:
            self.args.crawler = crawler

    @property
    def start_crawler(self) -> bool:
        return self.args.start_crawler

    @start_crawler.setter
    def start_crawler(self, start_crawler: bool) -> None:
        if start_crawler is not None:
            self.args.start_crawler = start_crawler


def create_glue_crawler(**kwargs) -> bool:

    args: CreateGlueCrawlerArgs = CreateGlueCrawlerArgs(**kwargs)
    # logger.debug("CreateGlueCrawlerArgs: {}".format(args))

    create_success = args.crawler.create_crawler()
    if create_success:
        logger.info("Create GlueCrawler: success")
        if args.start_crawler:
            logger.info(f"Starting GlueCrawler")
            start_success = args.crawler.start_crawler()
            if start_success:
                logger.info("GlueCrawler started")
            return start_success
    return create_success



from datetime import datetime
from typing import Optional, List, Dict, Any

from phidata.product import DataProduct, DataProductArgs
from phidata.types.run_status import RunStatus
from phidata.workflow import Workflow
from phidata.utils.env_var import validate_env_vars
from phidata.utils.log import logger
from phidata.utils.cli_console import print_info, print_subheading, print_error


class WorkflowChainArgs(DataProductArgs):
    workflows: List[Workflow]


class WorkflowChain(DataProduct):
    def __init__(
        self,
        name: str,
        workflows: List[Workflow],
        # Optional: airflow dag_id for this WorkflowChain
        dag_id: Optional[str] = None,
        validate_env: Optional[Dict[str, Any]] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ) -> None:

        super().__init__()
        try:
            self.args: WorkflowChainArgs = WorkflowChainArgs(
                name=name,
                workflows=workflows,
                dag_id=dag_id,
                validate_env=validate_env,
                version=version,
                enabled=enabled,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def workflows(self) -> Optional[List[Workflow]]:
        return self.args.workflows if self.args else None

    def run_in_local_env(self) -> bool:

        logger.debug("--**-- Running WorkflowChain locally")

        wf_run_status: List[RunStatus] = []
        for count, wf in enumerate(self.workflows, start=1):
            wf_name = wf.name or "{}__{}".format(wf.__class__.__name__, count)
            print_subheading(f"\nRunning {wf_name}")
            # logger.debug("Setting context")
            # Pass down context
            wf.run_context = self.run_context
            wf.path_context = self.path_context
            # Not required for local run but adding for posterity
            if self.dag_id:
                wf.dag_id = self.dag_id
            run_success = wf.run_in_local_env()
            wf_run_status.append(RunStatus(wf_name, run_success))

        print_subheading("\nWorkflow run status:")
        print_info(
            "\n".join(
                [
                    "{}: {}".format(wf.name, "Success" if wf.success else "Fail")
                    for wf in wf_run_status
                ]
            )
        )
        print_info("")
        for _run in wf_run_status:
            if not _run.success:
                return False
        return True

    def run_in_docker_container(self, active_container: Any) -> bool:

        logger.debug("--**-- Running WorkflowChain in docker container")

        wf_run_status: List[RunStatus] = []
        for count, wf in enumerate(self.workflows, start=1):
            wf_name = wf.name or "{}__{}".format(wf.__class__.__name__, count)
            print_subheading(f"\nRunning {wf_name}")
            # Pass down context
            wf.run_context = self.run_context
            wf.path_context = self.path_context
            wf.dag_id = self.dag_id
            run_success = wf.run_in_docker_container(active_container)
            wf_run_status.append(RunStatus(wf_name, run_success))

        print_subheading("\nWorkflow run status:")
        print_info(
            "\n".join(
                [
                    "{}: {}".format(wf.name, "Success" if wf.success else "Fail")
                    for wf in wf_run_status
                ]
            )
        )
        print_info("")
        for _run in wf_run_status:
            if not _run.success:
                return False
        return True

    def run_in_k8s_container(self) -> bool:
        logger.debug(f"@run_k8s not defined for {self.__class__.__name__}")
        return False

    def create_airflow_dag(
        self,
        owner: Optional[str] = "airflow",
        depends_on_past: Optional[bool] = False,
        # The description for the DAG to e.g. be shown on the webserver
        description: Optional[str] = None,
        # Defines how often that DAG runs, this
        #  timedelta object gets added to your latest task instance's
        #  execution_date to figure out the next schedule
        schedule_interval: Optional[Any] = None,
        # The timestamp from which the scheduler will
        #  attempt to backfill
        start_date: Optional[datetime] = None,
        # A date beyond which your DAG won't run, leave to None
        #  for open ended scheduling
        end_date: Optional[datetime] = None,
        # a dictionary of macros that will be exposed
        #  in your jinja templates. For example, passing ``dict(foo='bar')``
        #  to this argument allows you to ``{{ foo }}`` in all jinja
        #  templates related to this DAG. Note that you can pass any
        #  type of object here.
        user_defined_macros: Optional[Dict] = None,
        # a dictionary of filters that will be exposed
        #  in your jinja templates. For example, passing
        #  ``dict(hello=lambda name: 'Hello %s' % name)`` to this argument allows
        #  you to ``{{ 'world' | hello }}`` in all jinja templates related to
        #  this DAG.
        user_defined_filters: Optional[Dict] = None,
        # A dictionary of default parameters to be used
        #  as constructor keyword parameters when initialising operators.
        #  Note that operators have the same hook, and precede those defined
        #  here, meaning that if your dict contains `'depends_on_past': True`
        #  here and `'depends_on_past': False` in the operator's call
        #  `default_args`, the actual value will be `False`.
        default_args: Optional[Dict] = None,
        # a dictionary of DAG level parameters that are made
        # accessible in templates, namespaced under `params`. These
        # params can be overridden at the task level.
        params: Optional[Dict] = None,
        # the number of task instances allowed to run concurrently
        concurrency: Optional[int] = None,
        # maximum number of active DAG runs, beyond this
        #  number of DAG runs in a running state, the scheduler won't create
        #  new active DAG runs
        max_active_runs: Optional[int] = None,
        doc_md: Optional[str] = None,
        is_paused_upon_creation: Optional[bool] = None,
        jinja_environment_kwargs: Optional[Dict] = None,
        tags: Optional[List[str]] = None,
    ) -> Any:
        """
        Used to create an Airflow DAG for this WorkflowChain.
        """

        if not validate_env_vars(self.args.validate_airflow_env):
            # This function skips DAG creation on local machines
            # print_error(f"Could not validate airflow env: {self.args.validate_airflow_env}")
            return None

        from airflow import DAG
        from airflow.utils.dates import days_ago

        default_args = {
            "owner": owner,
            "depends_on_past": depends_on_past,
        }
        if start_date is None:
            start_date = days_ago(3)

        dag_id = self.dag_id
        if dag_id is None:
            print_error("Workflow dag_id unavailable")
            return False
        dag = DAG(
            self.dag_id if self.dag_id else "default_dag_id",
            default_args=default_args,
            description=description,
            schedule_interval=schedule_interval,
            start_date=start_date,
            end_date=end_date,
        )
        for workflow in self.workflows:
            _add_task_success = workflow.add_airflow_tasks_to_dag(dag)
            if not _add_task_success:
                print_error(
                    "Tasks for workflow {} could not be added".format(workflow.name)
                )
        print_info(f"Airflow dag {dag_id} created")
        return dag



import logging
import sys


def get_simple_logger(logger_name: str) -> logging.Logger:

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    simple_formatter = logging.Formatter(
        "%(lineno)d - %(filename)s - %(levelname)s - %(message)s"
    )
    # dttm_formatter = logging.Formatter(
    #     "%(asctime)s - %(lineno)d - %(filename)s - %(levelname)s - %(message)s",
    #     "%Y-%m-%d %H:%M",
    # )
    handler.setFormatter(simple_formatter)

    _logger = logging.getLogger(logger_name)
    _logger.addHandler(handler)
    _logger.setLevel(logging.DEBUG)
    _logger.propagate = False
    return _logger


def get_rich_logger(logger_name: str, show_path: bool = False) -> logging.Logger:

    from rich.logging import RichHandler

    # https://rich.readthedocs.io/en/latest/reference/logging.html#rich.logging.RichHandler
    # https://rich.readthedocs.io/en/latest/logging.html#handle-exceptions
    rich_handler = RichHandler(
        rich_tracebacks=False, show_time=False, show_path=show_path
    )
    rich_handler.setFormatter(
        logging.Formatter(
            fmt="%(message)s",
            datefmt="[%X]",
        )
    )

    _logger = logging.getLogger(logger_name)
    _logger.addHandler(rich_handler)
    _logger.setLevel(logging.INFO)
    _logger.propagate = False
    return _logger


logger: logging.Logger = get_rich_logger("phidata")



from typing import Optional
from typing_extensions import Literal

from phidata.infra.base import InfraConfig
from phidata.infra.aws.args import AwsArgs
from phidata.infra.aws.manager import AwsManager
from phidata.infra.aws.resource.group import AwsResourceGroup


class AwsConfig(InfraConfig):
    def __init__(
        self,
        name: Optional[str] = None,
        env: Optional[Literal["dev", "stg", "prd"]] = "prd",
        version: Optional[str] = None,
        enabled: bool = True,
        resources: Optional[AwsResourceGroup] = None,
        resources_dir: str = "aws",
        # AWS configuration for this env
        # Use aws config from WorkspaceConfig if not provided
        aws_region: Optional[str] = None,
        aws_profile: Optional[str] = None,
        aws_config_file: Optional[str] = None,
        aws_shared_credentials_file: Optional[str] = None,
    ):
        super().__init__()
        try:
            self.args: AwsArgs = AwsArgs(
                name=name,
                env=env,
                version=version,
                enabled=enabled,
                resources=resources,
                resources_dir=resources_dir,
                aws_region=aws_region,
                aws_profile=aws_profile,
                aws_config_file=aws_config_file,
                aws_shared_credentials_file=aws_shared_credentials_file,
            )
        except Exception as e:
            raise

    @property
    def resources(self) -> Optional[AwsResourceGroup]:
        return self.args.resources if self.args else None

    @property
    def resources_dir(self) -> Optional[str]:
        return self.args.resources_dir if self.args else None

    @resources_dir.setter
    def resources_dir(self, resources_dir: str) -> None:
        if self.args is not None and resources_dir is not None:
            self.args.resources_dir = resources_dir

    @property
    def aws_region(self) -> Optional[str]:
        return self.args.aws_region if self.args else None

    @aws_region.setter
    def aws_region(self, aws_region: str) -> None:
        if self.args is not None and aws_region is not None:
            self.args.aws_region = aws_region

    @property
    def aws_profile(self) -> Optional[str]:
        return self.args.aws_profile if self.args else None

    @aws_profile.setter
    def aws_profile(self, aws_profile: str) -> None:
        if self.args is not None and aws_profile is not None:
            self.args.aws_profile = aws_profile

    @property
    def aws_config_file(self) -> Optional[str]:
        return self.args.aws_config_file if self.args else None

    @aws_config_file.setter
    def aws_config_file(self, aws_config_file: str) -> None:
        if self.args is not None and aws_config_file is not None:
            self.args.aws_config_file = aws_config_file

    @property
    def aws_shared_credentials_file(self) -> Optional[str]:
        return self.args.aws_shared_credentials_file if self.args else None

    @aws_shared_credentials_file.setter
    def aws_shared_credentials_file(self, aws_shared_credentials_file: str) -> None:
        if self.args is not None and aws_shared_credentials_file is not None:
            self.args.aws_shared_credentials_file = aws_shared_credentials_file

    def resources_are_valid(self) -> bool:
        if self.resources is None:
            return True
        if not isinstance(self.resources, AwsResourceGroup):
            raise TypeError("Invalid AwsResourceGroup: {}".format(self.resources))
        return True

    def is_valid(self) -> bool:
        return self.resources_are_valid()

    def get_aws_manager(self) -> Optional[AwsManager]:
        return AwsManager(aws_args=self.args)



from typing import List, Optional, Dict

from pydantic import BaseModel

from phidata.infra.k8s.resource.apiextensions_k8s_io.v1.custom_object import (
    CustomObject,
)
from phidata.infra.k8s.resource.apiextensions_k8s_io.v1.custom_resource_definition import (
    CustomResourceDefinition,
)
from phidata.infra.k8s.resource.apps.v1.deployment import Deployment
from phidata.infra.k8s.resource.core.v1.config_map import ConfigMap
from phidata.infra.k8s.resource.core.v1.namespace import Namespace
from phidata.infra.k8s.resource.core.v1.secret import Secret
from phidata.infra.k8s.resource.core.v1.service import Service
from phidata.infra.k8s.resource.core.v1.service_account import ServiceAccount
from phidata.infra.k8s.resource.rbac_authorization_k8s_io.v1.cluste_role_binding import (
    ClusterRoleBinding,
)
from phidata.infra.k8s.resource.rbac_authorization_k8s_io.v1.cluster_role import (
    ClusterRole,
)
from phidata.infra.k8s.resource.storage_k8s_io.v1.storage_class import StorageClass


class K8sResourceGroup(BaseModel):
    """The K8sResourceGroup class contains the instructions to manage K8s resources"""

    name: str
    enabled: bool = True

    # The weight variable controls how this is resource group is deployed to a cluster
    # relative to other resource groups.

    # Within each resource group, different types of resources are
    # deployed in a predefined order (eg: ns first, then sa and so on..)
    # but we can also add an order to how to deploy different resource groups.
    # (Eg if we want to deploy a resource group with just storage_class (s) before all other resources)

    # Weights 1-10 are reserved
    # Weight 100 is default.
    # ResourceGroups with weight 100 are the default resources.
    # Choose weight 11-99 to deploy a resource group before all the default resources.
    # Choose weight 101+ to deploy a resource group after all the default resources
    weight: int = 100
    ns: Optional[Namespace] = None
    sa: Optional[ServiceAccount] = None
    cr: Optional[ClusterRole] = None
    crb: Optional[ClusterRoleBinding] = None
    secrets: Optional[List[Secret]] = None
    config_maps: Optional[List[ConfigMap]] = None
    storage_classes: Optional[List[StorageClass]] = None
    services: Optional[List[Service]] = None
    deployments: Optional[List[Deployment]] = None
    custom_objects: Optional[List[CustomObject]] = None
    crds: Optional[List[CustomResourceDefinition]] = None


class K8sBuildContext(BaseModel):
    namespace: str = "default"
    context: Optional[str] = None
    service_account_name: Optional[str] = None
    labels: Optional[Dict[str, str]] = None



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, List, Union
from typing_extensions import Literal

from phidata.app import PhidataApp, PhidataAppArgs
from phidata.app.traefik.crds import (
    ingressroute_crd,
    ingressroutetcp_crd,
    ingressrouteudp_crd,
    middleware_crd,
    middlewaretcp_crd,
    serverstransport_crd,
    tlsoption_crd,
    tlsstore_crd,
    traefikservice_crd,
)
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.k8s.create.apiextensions_k8s_io.v1.custom_object import (
    CreateCustomObject,
)
from phidata.infra.k8s.create.core.v1.service_account import CreateServiceAccount
from phidata.infra.k8s.create.core.v1.container import CreateContainer
from phidata.infra.k8s.create.core.v1.config_map import CreateConfigMap
from phidata.infra.k8s.create.apps.v1.deployment import CreateDeployment
from phidata.infra.k8s.create.core.v1.secret import CreateSecret
from phidata.infra.k8s.create.core.v1.service import CreateService, ServiceType
from phidata.infra.k8s.create.common.port import CreatePort
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.create.rbac_authorization_k8s_io.v1.cluster_role import (
    CreateClusterRole,
    PolicyRule,
)
from phidata.infra.k8s.create.rbac_authorization_k8s_io.v1.cluste_role_binding import (
    CreateClusterRoleBinding,
)
from phidata.infra.k8s.enums.image_pull_policy import ImagePullPolicy
from phidata.infra.k8s.enums.restart_policy import RestartPolicy
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.common import (
    get_default_service_name,
    get_default_container_name,
    get_default_deploy_name,
    get_default_pod_name,
    get_default_cr_name,
    get_default_configmap_name,
    get_default_crb_name,
    get_default_sa_name,
    get_default_secret_name,
)
from phidata.utils.enums import ExtendedEnum
from phidata.utils.cli_console import print_error, print_info, print_warning
from phidata.utils.log import logger


class LoadBalancerProvider(ExtendedEnum):
    AWS = "AWS"


class IngressRouteArgs(PhidataAppArgs):
    name: str = "ingressroute"
    version: str = "1"
    enabled: bool = True

    # Image Args
    image_name: str = "traefik"
    image_tag: str = "v2.6"

    domain_name: Optional[str] = None

    # Configure traefik
    web_enabled: bool = False
    web_routes: Optional[List[dict]] = None
    web_container_port: int = 80
    web_service_port: int = 80
    web_node_port: Optional[int] = None
    web_key: str = "web"
    web_ingress_name: str = "web-ingress"
    forward_web_to_websecure: bool = False

    websecure_enabled: bool = False
    websecure_routes: Optional[List[dict]] = None
    websecure_container_port: int = 443
    websecure_service_port: int = 443
    websecure_node_port: Optional[int] = None
    websecure_key: str = "websecure"
    websecure_ingress_name: str = "websecure-ingress"

    dashboard_enabled: bool = (False,)
    dashboard_routes: Optional[List[dict]] = None
    dashboard_container_port: int = 8080
    dashboard_service_port: int = 8080
    dashboard_node_port: Optional[int] = None
    dashboard_key: str = "dashboard"
    dashboard_ingress_name: str = "dashboard-ingress"
    # The dashboard is gated behind a user:password, which is generated using
    #   htpasswd -nb user password
    # You can provide the "users:password" list as a dashboard_auth_users param
    # or as DASHBOARD_AUTH_USERS in the secrets_file
    # Using the secrets_file is recommended
    dashboard_auth_users: Optional[str] = None
    insecure_api_access: bool = False

    # Traefik config
    # Enable Access Logs
    access_logs: bool = True
    # Traefik config file on the host
    traefik_config_file: Optional[str] = None
    # Traefik config file on the container
    traefik_config_file_container_path: Path = Path("/etc/traefik/traefik.yaml")

    # Configure rbac
    sa_name: Optional[str] = None
    cr_name: Optional[str] = None
    crb_name: Optional[str] = None
    # Configure the container
    container_name: Optional[str] = None
    container_args: Optional[List[str]] = None
    image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT
    # Configure the deployment
    deploy_name: Optional[str] = None
    pod_name: Optional[str] = None
    replicas: int = 1
    pod_node_selector: Optional[Dict[str, str]] = None
    restart_policy: RestartPolicy = RestartPolicy.ALWAYS
    termination_grace_period_seconds: Optional[int] = None
    # Configure the service
    service_name: Optional[str] = None
    service_type: ServiceType = ServiceType.LOAD_BALANCER
    service_annotations: Optional[Dict[str, Optional[str]]] = None

    # Add env variables to container env
    env: Optional[Dict[str, str]] = None
    # Read env variables from a file in yaml format
    env_file: Optional[Path] = None
    # Configure the ConfigMap used for env variables that are not Secret
    config_map_name: Optional[str] = None
    # Configure the Secret used for env variables that are Secret
    secret_name: Optional[str] = None
    # Read secrets from a file in yaml format
    secrets_file: Optional[Path] = None

    # On cloud providers which support external load balancers,
    # setting the service_type field to LoadBalancer provisions a load balancer.
    # The actual creation of the load balancer happens asynchronously
    #
    # load_balancer_provider is required if service_type == ServiceType.LOAD_BALANCER
    load_balancer_provider: Optional[LoadBalancerProvider] = None

    # AWS LoadBalancer configuration
    use_nlb: bool = True
    # Specifies the target type to configure for NLB. You can choose between instance and ip.
    # `instance` mode will route traffic to all EC2 instances within cluster on the NodePort opened for your service.
    #       service must be of type NodePort or LoadBalancer for instance targets
    #       for k8s 1.22 and later if spec.allocateLoadBalancerNodePorts is set to false,
    #       NodePort must be allocated manually
    # `ip` mode will route traffic directly to the pod IP.
    #       network plugin must use native AWS VPC networking configuration for pod IP,
    #       for example Amazon VPC CNI plugin.
    nlb_target_type: Optional[Literal["instance", "ip"]] = None
    # Write Access Logs to s3
    access_logs_to_s3: bool = False
    # The name of the aws S3 bucket where the access logs are stored
    access_logs_s3_bucket: Optional[str] = None
    # The logical hierarchy you created for your aws S3 bucket, for example `my-bucket-prefix/prod`
    access_logs_s3_bucket_prefix: Optional[str] = None
    # If provided, TLS termination is added to the LB
    acm_certificate_arn: Optional[str] = None
    acm_certificate_summary_file: Optional[Path] = (None,)
    load_balancer_ip: Optional[str] = None
    # If None, default is internal.
    load_balancer_scheme: Optional[Literal["internal", "internet-facing"]] = None
    # Limit the IPs that can access this endpoint
    # You can provide the load_balancer_source_ranges as a list here
    # or as LOAD_BALANCER_SOURCE_RANGES in the secrets_file
    # Using the secrets_file is recommended
    load_balancer_source_ranges: Optional[List[str]] = None
    allocate_load_balancer_node_ports: Optional[bool] = None


class IngressRoute(PhidataApp):
    def __init__(
        self,
        name: str = "ingressroute",
        version: str = "1",
        enabled: bool = True,
        # Image Args,
        image_name: str = "traefik",
        image_tag: str = "v2.6",
        domain_name: Optional[str] = None,
        # Configure traefik,
        web_enabled: bool = False,
        web_routes: Optional[List[dict]] = None,
        web_container_port: int = 80,
        web_service_port: int = 80,
        web_node_port: Optional[int] = None,
        web_key: str = "web",
        web_ingress_name: str = "web-ingress",
        forward_web_to_websecure: bool = False,
        websecure_enabled: bool = False,
        websecure_routes: Optional[List[dict]] = None,
        websecure_container_port: int = 443,
        websecure_service_port: int = 443,
        websecure_node_port: Optional[int] = None,
        websecure_key: str = "websecure",
        websecure_ingress_name: str = "websecure-ingress",
        dashboard_enabled: bool = (False,),
        dashboard_routes: Optional[List[dict]] = None,
        dashboard_container_port: int = 8080,
        dashboard_service_port: int = 8080,
        dashboard_node_port: Optional[int] = None,
        dashboard_key: str = "dashboard",
        dashboard_ingress_name: str = "dashboard-ingress",
        # The dashboard is gated behind a user:password, which is generated using
        #   htpasswd -nb user password
        # You can provide the "users:password" list as a dashboard_auth_users param
        # or as DASHBOARD_AUTH_USERS in the secrets_file
        # Using the secrets_file is recommended
        dashboard_auth_users: Optional[str] = None,
        insecure_api_access: bool = False,
        # Traefik config,
        # Enable Access Logs,
        access_logs: bool = True,
        # Traefik config file on the host,
        traefik_config_file: Optional[str] = None,
        # Traefik config file on the container,
        traefik_config_file_container_path: Path = Path("/etc/traefik/traefik.yaml"),
        # Configure rbac,
        sa_name: Optional[str] = None,
        cr_name: Optional[str] = None,
        crb_name: Optional[str] = None,
        # Configure the container,
        container_name: Optional[str] = None,
        container_args: Optional[List[str]] = None,
        image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT,
        # Configure the deployment,
        deploy_name: Optional[str] = None,
        pod_name: Optional[str] = None,
        replicas: int = 1,
        pod_node_selector: Optional[Dict[str, str]] = None,
        restart_policy: RestartPolicy = RestartPolicy.ALWAYS,
        termination_grace_period_seconds: Optional[int] = None,
        # Configure the service,
        service_name: Optional[str] = None,
        service_type: ServiceType = ServiceType.LOAD_BALANCER,
        service_annotations: Optional[Dict[str, Optional[str]]] = None,
        # Add env variables to container env,
        env: Optional[Dict[str, str]] = None,
        # Read env variables from a file in yaml format,
        env_file: Optional[Path] = None,
        # Configure the ConfigMap used for env variables that are not Secret,
        config_map_name: Optional[str] = None,
        # Configure the Secret used for env variables that are Secret,
        secret_name: Optional[str] = None,
        # Read secrets from a file in yaml format,
        secrets_file: Optional[Path] = None,
        # On cloud providers which support external load balancers,,
        # setting the service_type field to LoadBalancer provisions a load balancer.,
        # The actual creation of the load balancer happens asynchronously,
        # load_balancer_provider is required if service_type == ServiceType.LOAD_BALANCER,
        load_balancer_provider: Optional[LoadBalancerProvider] = None,
        # AWS LoadBalancer configuration,
        use_nlb: bool = True,
        # Specifies the target type to configure for NLB. You can choose between instance and ip.,
        # `instance` mode will route traffic to all EC2 instances within cluster on the NodePort opened for your service.,
        #       service must be of type NodePort or LoadBalancer for instance targets,
        #       for k8s 1.22 and later if spec.allocateLoadBalancerNodePorts is set to false,,
        #       NodePort must be allocated manually,
        # `ip` mode will route traffic directly to the pod IP.,
        #       network plugin must use native AWS VPC networking configuration for pod IP,,
        #       for example Amazon VPC CNI plugin.,
        nlb_target_type: Optional[Literal["instance", "ip"]] = None,
        # Write Access Logs to s3,
        access_logs_to_s3: bool = False,
        # The name of the aws S3 bucket where the access logs are stored,
        access_logs_s3_bucket: Optional[str] = None,
        # The logical hierarchy you created for your aws S3 bucket, for example `my-bucket-prefix/prod`,
        access_logs_s3_bucket_prefix: Optional[str] = None,
        # If provided, TLS termination is added to the LB,
        acm_certificate_arn: Optional[str] = None,
        acm_certificate_summary_file: Optional[Path] = (None,),
        load_balancer_ip: Optional[str] = None,
        # If None, default is internal.,
        load_balancer_scheme: Optional[Literal["internal", "internet-facing"]] = None,
        # Limit the IPs that can access this endpoint
        # You can provide the load_balancer_source_ranges as a list here
        # or as LOAD_BALANCER_SOURCE_RANGES in the secrets_file
        # Using the secrets_file is recommended
        load_balancer_source_ranges: Optional[List[str]] = None,
        allocate_load_balancer_node_ports: Optional[bool] = None,
        # Additional args
        # If True, skip resource creation if active resources with the same name exist.
        use_cache: bool = True,
        # If True, log extra debug messages
        use_verbose_logs: bool = False,
    ):
        super().__init__()
        try:
            self.args: IngressRouteArgs = IngressRouteArgs(
                name=name,
                version=version,
                enabled=enabled,
                image_name=image_name,
                image_tag=image_tag,
                domain_name=domain_name,
                web_enabled=web_enabled,
                web_routes=web_routes,
                web_container_port=web_container_port,
                web_service_port=web_service_port,
                web_node_port=web_node_port,
                web_key=web_key,
                web_ingress_name=web_ingress_name,
                forward_web_to_websecure=forward_web_to_websecure,
                websecure_enabled=websecure_enabled,
                websecure_routes=websecure_routes,
                websecure_container_port=websecure_container_port,
                websecure_service_port=websecure_service_port,
                websecure_node_port=websecure_node_port,
                websecure_key=websecure_key,
                websecure_ingress_name=websecure_ingress_name,
                dashboard_enabled=dashboard_enabled,
                dashboard_routes=dashboard_routes,
                dashboard_container_port=dashboard_container_port,
                dashboard_service_port=dashboard_service_port,
                dashboard_node_port=dashboard_node_port,
                dashboard_key=dashboard_key,
                dashboard_ingress_name=dashboard_ingress_name,
                dashboard_auth_users=dashboard_auth_users,
                insecure_api_access=insecure_api_access,
                access_logs=access_logs,
                traefik_config_file=traefik_config_file,
                traefik_config_file_container_path=traefik_config_file_container_path,
                sa_name=sa_name,
                cr_name=cr_name,
                crb_name=crb_name,
                container_name=container_name,
                container_args=container_args,
                image_pull_policy=image_pull_policy,
                deploy_name=deploy_name,
                pod_name=pod_name,
                replicas=replicas,
                pod_node_selector=pod_node_selector,
                restart_policy=restart_policy,
                termination_grace_period_seconds=termination_grace_period_seconds,
                service_name=service_name,
                service_type=service_type,
                service_annotations=service_annotations,
                env=env,
                env_file=env_file,
                config_map_name=config_map_name,
                secret_name=secret_name,
                secrets_file=secrets_file,
                load_balancer_provider=load_balancer_provider,
                use_nlb=use_nlb,
                nlb_target_type=nlb_target_type,
                access_logs_to_s3=access_logs_to_s3,
                access_logs_s3_bucket=access_logs_s3_bucket,
                access_logs_s3_bucket_prefix=access_logs_s3_bucket_prefix,
                acm_certificate_arn=acm_certificate_arn,
                acm_certificate_summary_file=acm_certificate_summary_file,
                load_balancer_ip=load_balancer_ip,
                load_balancer_scheme=load_balancer_scheme,
                load_balancer_source_ranges=load_balancer_source_ranges,
                allocate_load_balancer_node_ports=allocate_load_balancer_node_ports,
                use_cache=use_cache,
                use_verbose_logs=use_verbose_logs,
            )
        except Exception:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    def get_container_name(self) -> str:
        return self.args.container_name or get_default_container_name(self.args.name)

    def get_service_name(self) -> str:
        return self.args.service_name or get_default_service_name(self.args.name)

    def get_env_data_from_file(self) -> Optional[Dict[str, str]]:
        import yaml

        env_file_path = self.args.env_file
        if (
            env_file_path is not None
            and env_file_path.exists()
            and env_file_path.is_file()
        ):
            # logger.debug(f"Reading {env_file_path}")
            env_data_from_file = yaml.safe_load(env_file_path.read_text())
            if env_data_from_file is not None and isinstance(env_data_from_file, dict):
                return env_data_from_file
            else:
                print_error(f"Invalid env_file: {env_file_path}")
        return None

    def get_secret_data_from_file(self) -> Optional[Dict[str, str]]:
        import yaml

        secrets_file_path = self.args.secrets_file
        if (
            secrets_file_path is not None
            and secrets_file_path.exists()
            and secrets_file_path.is_file()
        ):
            # logger.debug(f"Reading {secrets_file_path}")
            secret_data_from_file = yaml.safe_load(secrets_file_path.read_text())
            if secret_data_from_file is not None and isinstance(
                secret_data_from_file, dict
            ):
                return secret_data_from_file
            else:
                print_error(f"Invalid secrets_file: {secrets_file_path}")
        return None

    def get_dashboard_auth_users(self) -> Optional[str]:
        dashboard_auth_users_var: Optional[str] = (
            self.args.dashboard_auth_users if self.args else None
        )
        if dashboard_auth_users_var is None and self.args.secrets_file is not None:
            # read from secrets_file
            logger.debug(f"Reading DASHBOARD_AUTH_USERS from secrets_file")
            secret_data_from_file = self.get_secret_data_from_file()
            if secret_data_from_file is not None:
                dashboard_auth_users_var = secret_data_from_file.get(
                    "DASHBOARD_AUTH_USERS", dashboard_auth_users_var
                )
        return dashboard_auth_users_var

    def get_load_balancer_source_ranges(self) -> Optional[List[str]]:
        load_balancer_source_ranges_var: Optional[List[str]] = (
            self.args.load_balancer_source_ranges if self.args else None
        )
        if (
            load_balancer_source_ranges_var is None
            and self.args.secrets_file is not None
        ):
            # read from secrets_file
            logger.debug(f"Reading LOAD_BALANCER_SOURCE_RANGES from secrets_file")
            secret_data_from_file = self.get_secret_data_from_file()
            if secret_data_from_file is not None:
                load_balancer_source_ranges_var = secret_data_from_file.get(
                    "LOAD_BALANCER_SOURCE_RANGES", load_balancer_source_ranges_var
                )
        return load_balancer_source_ranges_var

    ######################################################
    ## Docker Resources
    ######################################################

    def get_ingress_route_docker_rg(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[DockerResourceGroup]:
        print_error(f"IngressRoute not available on Docker")
        return None

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        docker_rg: Optional[DockerResourceGroup] = self.get_ingress_route_docker_rg(
            docker_build_context
        )
        # logger.debug("docker_rg:\n{}".format(docker_rg.json(indent=2)))
        if docker_rg is not None:
            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[docker_rg.name] = docker_rg

    ######################################################
    ## K8s Resources
    ######################################################

    def get_ingress_route_k8s_rg(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[K8sResourceGroup]:

        app_name = self.args.name
        logger.debug(f"Building {app_name} K8sResourceGroup")

        # Define K8s resources
        sa = CreateServiceAccount(
            sa_name=self.args.sa_name or get_default_sa_name(app_name),
            app_name=app_name,
            namespace=k8s_build_context.namespace,
        )

        cr = CreateClusterRole(
            cr_name=self.args.cr_name or get_default_cr_name(app_name),
            rules=[
                PolicyRule(
                    api_groups=[""],
                    resources=[
                        "services",
                        "endpoints",
                        "secrets",
                    ],
                    verbs=[
                        "get",
                        "list",
                        "watch",
                    ],
                ),
                PolicyRule(
                    api_groups=[
                        "extensions",
                        "networking.k8s.io",
                    ],
                    resources=[
                        "ingresses",
                        "ingressclasses",
                    ],
                    verbs=[
                        "get",
                        "list",
                        "watch",
                    ],
                ),
                PolicyRule(
                    api_groups=[
                        "extensions",
                    ],
                    resources=[
                        "ingresses/status",
                    ],
                    verbs=[
                        "update",
                    ],
                ),
                PolicyRule(
                    api_groups=[
                        "traefik.containo.us",
                    ],
                    resources=[
                        "middlewares",
                        "middlewaretcps",
                        "ingressroutes",
                        "traefikservices",
                        "ingressroutetcps",
                        "ingressrouteudps",
                        "tlsoptions",
                        "tlsstores",
                        "serverstransports",
                    ],
                    verbs=[
                        "get",
                        "list",
                        "watch",
                    ],
                ),
            ],
            app_name=app_name,
        )

        crb = CreateClusterRoleBinding(
            crb_name=get_default_crb_name(app_name),
            cr_name=cr.cr_name,
            service_account_name=sa.sa_name,
            app_name=app_name,
            namespace=k8s_build_context.namespace,
        )

        # Container Environment
        container_env: Dict[str, str] = {}
        # Update the container env using env_file
        env_data_from_file = self.get_env_data_from_file()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)
        # Update the container env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)
        # Create a ConfigMap to set the container env variables which are not Secret
        container_env_cm = CreateConfigMap(
            cm_name=self.args.config_map_name or get_default_configmap_name(app_name),
            app_name=app_name,
            data=container_env,
        )
        config_maps: List[CreateConfigMap] = [container_env_cm]

        # Create a Secret to set the container env variables which are Secret
        secrets: List[CreateSecret] = []
        secret_data_from_file = self.get_secret_data_from_file()
        if secret_data_from_file is not None:
            container_env_secret = CreateSecret(
                secret_name=self.args.secret_name or get_default_secret_name(app_name),
                app_name=app_name,
                string_data=secret_data_from_file,
            )
            secrets.append(container_env_secret)

        custom_objects: List[CreateCustomObject] = []
        ports: List[CreatePort] = []

        container_args = self.args.container_args
        if container_args is None:
            container_args = []
        container_args.append("--providers.kubernetescrd")

        if self.args.access_logs:
            container_args.append("--accesslog")

        if self.args.web_enabled:
            container_args.append(
                f"--entrypoints.{self.args.web_key}.Address=:{self.args.web_service_port}"
            )

            web_port = CreatePort(
                name=self.args.web_key,
                container_port=self.args.web_container_port,
                service_port=self.args.web_service_port,
                target_port=self.args.web_key,
            )
            if (
                self.args.service_type
                in (ServiceType.NODE_PORT, ServiceType.LOAD_BALANCER)
                and self.args.web_node_port is not None
            ):
                web_port.node_port = self.args.web_node_port
            ports.append(web_port)

            web_ingressroute = CreateCustomObject(
                name=self.args.web_ingress_name,
                crd=ingressroute_crd,
                spec={
                    "entryPoints": [self.args.web_key],
                    "routes": self.args.web_routes,
                },
                app_name=app_name,
                namespace=k8s_build_context.namespace,
            )
            custom_objects.append(web_ingressroute)

        if self.args.websecure_enabled:
            container_args.append(
                f"--entrypoints.{self.args.websecure_key}.Address=:{self.args.websecure_service_port}"
            )
            if self.args.forward_web_to_websecure:
                container_args.extend(
                    [
                        f"--entrypoints.{self.args.web_key}.http.redirections.entryPoint.to={self.args.websecure_key}",
                        f"--entrypoints.{self.args.web_key}.http.redirections.entryPoint.scheme=https",
                    ]
                )

            websecure_port = CreatePort(
                name=self.args.websecure_key,
                container_port=self.args.websecure_container_port,
                service_port=self.args.websecure_service_port,
                target_port=self.args.websecure_key,
            )
            if (
                self.args.service_type
                in (ServiceType.NODE_PORT, ServiceType.LOAD_BALANCER)
                and self.args.websecure_node_port is not None
            ):
                websecure_port.node_port = self.args.websecure_node_port
            ports.append(websecure_port)

            websecure_ingressroute = CreateCustomObject(
                name=self.args.websecure_ingress_name,
                crd=ingressroute_crd,
                spec={
                    "entryPoints": [self.args.websecure_key],
                    "routes": self.args.websecure_routes,
                },
                app_name=app_name,
                namespace=k8s_build_context.namespace,
            )
            custom_objects.append(websecure_ingressroute)

        if self.args.dashboard_enabled:
            container_args.append(f"--api=true")
            container_args.append(f"--api.dashboard=true")
            if self.args.insecure_api_access:
                container_args.append(f"--api.insecure")

            dashboard_port = CreatePort(
                name=self.args.dashboard_key,
                container_port=self.args.dashboard_container_port,
                service_port=self.args.dashboard_service_port,
                target_port=self.args.dashboard_key,
            )
            if (
                self.args.service_type
                in (ServiceType.NODE_PORT, ServiceType.LOAD_BALANCER)
                and self.args.dashboard_node_port is not None
            ):
                dashboard_port.node_port = self.args.dashboard_node_port
            ports.append(dashboard_port)

            # create dashboard_auth_middleware if auth provided
            # ref: https://doc.traefik.io/traefik/operations/api/#configuration
            dashboard_auth_middleware = None
            dashboard_auth_users = self.get_dashboard_auth_users()
            if dashboard_auth_users is not None:
                dashboard_auth_secret = CreateSecret(
                    secret_name="dashboard-auth-secret",
                    app_name=app_name,
                    namespace=k8s_build_context.namespace,
                    data={"users": dashboard_auth_users},
                )
                secrets.append(dashboard_auth_secret)

                dashboard_auth_middleware = CreateCustomObject(
                    name="dashboard-auth-middleware",
                    crd=middleware_crd,
                    spec={"basicAuth": {"secret": dashboard_auth_secret.secret_name}},
                    app_name=app_name,
                    namespace=k8s_build_context.namespace,
                )
                custom_objects.append(dashboard_auth_middleware)

            dashboard_routes = self.args.dashboard_routes
            # use default dashboard routes
            if dashboard_routes is None:
                # domain must be provided
                if self.args.domain_name is not None:
                    dashboard_routes = [
                        {
                            "kind": "Rule",
                            "match": f"Host(`traefik.{self.args.domain_name}`)",
                            "middlewares": [
                                {
                                    "name": dashboard_auth_middleware.name,
                                    "namespace": k8s_build_context.namespace,
                                },
                            ]
                            if dashboard_auth_middleware is not None
                            else [],
                            "services": [
                                {
                                    "kind": "TraefikService",
                                    "name": "api@internal",
                                }
                            ],
                        },
                    ]

            # TODO: uncomment if need to add dashboard_entrypoint
            # container_args.append(
            #     f"--entrypoints.{self.args.dashboard_key}.Address=:{self.args.dashboard_service_port}"
            # )
            # # dashboard_entrypoint: default to dashboard_key i.e. 8080
            # dashboard_entrypoint = self.args.dashboard_key
            # # if web/http at 80 is avl, use that
            # if self.args.web_enabled:
            #     dashboard_entrypoint = self.args.web_key
            # # if websecure/https at 443 is avl, then use that over web/http
            # if self.args.websecure_enabled:
            #     dashboard_entrypoint = self.args.websecure_key

            dashboard_ingressroute = CreateCustomObject(
                name=self.args.dashboard_ingress_name,
                crd=ingressroute_crd,
                spec={
                    # "entryPoints": [dashboard_entrypoint],
                    "routes": dashboard_routes,
                },
                app_name=app_name,
                namespace=k8s_build_context.namespace,
            )
            custom_objects.append(dashboard_ingressroute)

        container = CreateContainer(
            container_name=self.get_container_name(),
            app_name=app_name,
            image_name=self.args.image_name,
            image_tag=self.args.image_tag,
            args=container_args,
            ports=ports,
        )

        deployment = CreateDeployment(
            deploy_name=self.args.deploy_name or get_default_deploy_name(app_name),
            pod_name=self.args.pod_name or get_default_pod_name(app_name),
            app_name=app_name,
            namespace=k8s_build_context.namespace,
            service_account_name=sa.sa_name,
            containers=[container],
            labels=k8s_build_context.labels,
        )

        service_annotations = self.args.service_annotations
        if service_annotations is None:
            service_annotations = OrderedDict()

        # Configure AWS LoadBalancer
        if self.args.load_balancer_provider == LoadBalancerProvider.AWS:
            # https://kubernetes.io/docs/concepts/services-networking/service/#aws-nlb-support
            if self.args.use_nlb:
                service_annotations[
                    "service.beta.kubernetes.io/aws-load-balancer-type"
                ] = "nlb"
            if self.args.nlb_target_type is not None:
                service_annotations[
                    "service.beta.kubernetes.io/aws-load-balancer-nlb-target-type"
                ] = self.args.nlb_target_type

            # New: https://kubernetes-sigs.github.io/aws-load-balancer-controller/v2.4/guide/service/annotations/#load-balancer-attributes
            # Deprecated docs: # https://kubernetes.io/docs/concepts/services-networking/service/#elb-access-logs-on-aws
            if self.args.access_logs_to_s3:
                lb_attributes = ""
                lb_attributes += "access_logs.s3.enabled=true"
                if self.args.access_logs_s3_bucket is not None:
                    lb_attributes += (
                        f",access_logs.s3.bucket={self.args.access_logs_s3_bucket}"
                    )
                if self.args.access_logs_s3_bucket_prefix is not None:
                    lb_attributes += f",access_logs.s3.prefix={self.args.access_logs_s3_bucket_prefix}"
                service_annotations[
                    "service.beta.kubernetes.io/aws-load-balancer-attributes"
                ] = lb_attributes

            # https://kubernetes-sigs.github.io/aws-load-balancer-controller/v2.4/guide/service/annotations/#ssl-cert
            if self.args.acm_certificate_arn is not None:
                service_annotations[
                    "service.beta.kubernetes.io/aws-load-balancer-ssl-cert"
                ] = self.args.acm_certificate_arn
                service_annotations[
                    "service.beta.kubernetes.io/aws-load-balancer-ssl-ports"
                ] = str(self.args.websecure_service_port)
            # if acm_certificate_summary_file is provided, use that
            if self.args.acm_certificate_summary_file is not None and isinstance(
                self.args.acm_certificate_summary_file, Path
            ):
                if (
                    self.args.acm_certificate_summary_file.exists()
                    and self.args.acm_certificate_summary_file.is_file()
                ):
                    from phidata.infra.aws.resource.acm.certificate import (
                        CertificateSummary,
                    )

                    cert_summary = CertificateSummary.parse_file(
                        self.args.acm_certificate_summary_file
                    )
                    certificate_arn = cert_summary.CertificateArn
                    logger.debug(f"CertificateArn: {certificate_arn}")
                    service_annotations[
                        "service.beta.kubernetes.io/aws-load-balancer-ssl-cert"
                    ] = certificate_arn
                    service_annotations[
                        "service.beta.kubernetes.io/aws-load-balancer-ssl-ports"
                    ] = str(self.args.websecure_service_port)
                else:
                    print_warning(
                        f"Does not exist: {self.args.acm_certificate_summary_file}"
                    )

            if self.args.load_balancer_scheme is not None:
                service_annotations[
                    "sservice.beta.kubernetes.io/aws-load-balancer-scheme"
                ] = self.args.load_balancer_scheme

        service = CreateService(
            service_name=self.get_service_name(),
            app_name=app_name,
            namespace=k8s_build_context.namespace,
            service_account_name=sa.sa_name,
            service_type=self.args.service_type,
            deployment=deployment,
            ports=ports,
            labels=k8s_build_context.labels,
            annotations=service_annotations,
            load_balancer_ip=self.args.load_balancer_ip,
            load_balancer_source_ranges=self.get_load_balancer_source_ranges(),
            allocate_load_balancer_node_ports=self.args.allocate_load_balancer_node_ports,
        )

        resource_group = CreateK8sResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            sa=sa,
            cr=cr,
            crb=crb,
            crds=[
                ingressroute_crd,
                ingressroutetcp_crd,
                ingressrouteudp_crd,
                middleware_crd,
                middlewaretcp_crd,
                serverstransport_crd,
                tlsoption_crd,
                tlsstore_crd,
                traefikservice_crd,
            ],
            config_maps=config_maps,
            secrets=secrets,
            services=[service],
            deployments=[deployment],
            custom_objects=custom_objects,
        )

        return resource_group.create()

    def init_k8s_resource_groups(self, k8s_build_context: K8sBuildContext) -> None:
        k8s_rg: Optional[K8sResourceGroup] = self.get_ingress_route_k8s_rg(
            k8s_build_context
        )
        # logger.debug("k8s_rg:\n{}".format(k8s_rg.json(indent=2)))
        if k8s_rg is not None:
            if self.k8s_resource_groups is None:
                self.k8s_resource_groups = OrderedDict()
            self.k8s_resource_groups[k8s_rg.name] = k8s_rg



from datetime import datetime, timedelta
from pathlib import Path
from typing import Optional, Any, List, Dict

from pydantic import BaseModel

from phidata.constants import (
    SCRIPTS_DIR_ENV_VAR,
    STORAGE_DIR_ENV_VAR,
    META_DIR_ENV_VAR,
    PRODUCTS_DIR_ENV_VAR,
    NOTEBOOKS_DIR_ENV_VAR,
    WORKSPACE_CONFIG_DIR_ENV_VAR,
)
from phidata.utils.cli_console import print_info, print_error, print_subheading
from phidata.utils.log import logger
from phidata.types.run_status import RunStatus
from phidata.types.context import PathContext, RunContext
from phidata.workflow import Workflow
from phidata.utils.env_var import validate_env_vars


class DataProductArgs(BaseModel):
    name: str
    version: Optional[str] = None
    enabled: bool = True

    # List of workflows in this DataProduct
    workflows: List[Workflow]

    # Optional: airflow dag_id for this DataProduct, use name if not provided
    dag_id: Optional[str] = None
    # Optional: suffix to add to airflow tasks
    task_suffix: Optional[str] = None
    # run_context is provided by wf_operator
    run_context: Optional[RunContext] = None
    # path_context is provided using env variables
    path_context: Optional[PathContext] = None

    # Env vars that validate Airflow is active on the containers
    # These vars are useful in gating sections which should only run on
    # remote containers like creating DAGs and tasks
    validate_airflow_env: Dict[str, Any] = {"INIT_AIRFLOW": True}

    class Config:
        arbitrary_types_allowed = True


class DataProduct:
    """Base Class for all DataProducts"""

    def __init__(
        self,
        name: str,
        # List of workflows in this DataProduct
        workflows: List[Workflow],
        # Optional: airflow dag_id for this DataProduct
        dag_id: Optional[str] = None,
        # Optional: suffix to add to airflow tasks
        task_suffix: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ) -> None:
        try:
            self.args: DataProductArgs = DataProductArgs(
                name=name,
                workflows=workflows,
                dag_id=dag_id,
                task_suffix=task_suffix,
                version=version,
                enabled=enabled,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def name(self) -> str:
        return self.args.name if self.args else self.__class__.__name__

    @property
    def version(self) -> Optional[str]:
        return self.args.version if self.args else None

    @property
    def enabled(self) -> bool:
        return self.args.enabled if self.args else False

    @property
    def workflows(self) -> Optional[List[Workflow]]:
        return self.args.workflows if self.args else None

    @workflows.setter
    def workflows(self, workflows: List[Workflow]) -> None:
        if self.args is not None and workflows is not None:
            self.args.workflows = workflows

    @property
    def dag_id(self) -> str:
        return self.args.dag_id if (self.args and self.args.dag_id) else self.name

    @dag_id.setter
    def dag_id(self, dag_id: str) -> None:
        if self.args is not None and dag_id is not None:
            self.args.dag_id = dag_id

    @property
    def task_suffix(self) -> Optional[str]:
        return self.args.task_suffix if self.args else None

    @task_suffix.setter
    def task_suffix(self, task_suffix: str) -> None:
        if self.args is not None and task_suffix is not None:
            self.args.task_suffix = task_suffix

    @property
    def run_context(self) -> Optional[RunContext]:
        return self.args.run_context if self.args else None

    @run_context.setter
    def run_context(self, run_context: RunContext) -> None:
        if self.args is not None and run_context is not None:
            self.args.run_context = run_context

    @property
    def path_context(self) -> Optional[PathContext]:
        # DataProduct not yet initialized
        if self.args is None:
            return None

        if self.args.path_context is not None:
            # use cached value if available
            return self.args.path_context

        logger.debug(f"--++**++--> Loading PathContext from env")
        self.path_context = PathContext()

        import os

        scripts_dir = os.getenv(SCRIPTS_DIR_ENV_VAR)
        storage_dir = os.getenv(STORAGE_DIR_ENV_VAR)
        meta_dir = os.getenv(META_DIR_ENV_VAR)
        products_dir = os.getenv(PRODUCTS_DIR_ENV_VAR)
        notebooks_dir = os.getenv(NOTEBOOKS_DIR_ENV_VAR)
        workspace_config_dir = os.getenv(WORKSPACE_CONFIG_DIR_ENV_VAR)

        if storage_dir is None:
            print_error(f"{STORAGE_DIR_ENV_VAR} not set")
        if products_dir is None:
            print_error(f"{PRODUCTS_DIR_ENV_VAR} not set")

        try:
            if scripts_dir is not None:
                self.path_context.scripts_dir = Path(scripts_dir)
            if storage_dir is not None:
                self.path_context.storage_dir = Path(storage_dir)
            if meta_dir is not None:
                self.path_context.meta_dir = Path(meta_dir)
            if products_dir is not None:
                self.path_context.products_dir = Path(products_dir)
            if notebooks_dir is not None:
                self.path_context.notebooks_dir = Path(notebooks_dir)
            if workspace_config_dir is not None:
                self.path_context.workspace_config_dir = Path(workspace_config_dir)
        except Exception as e:
            raise
        logger.debug(f"--++**++--> PathContext loaded")
        return self.args.path_context

    @path_context.setter
    def path_context(self, path_context: PathContext) -> None:
        if self.args is not None and path_context is not None:
            self.args.path_context = path_context

    @property
    def validate_airflow_env(self) -> Dict[str, Any]:
        return self.args.validate_airflow_env if self.args else None

    @validate_airflow_env.setter
    def validate_airflow_env(self, validate_airflow_env: Dict[str, Any]) -> None:
        if self.args is not None and validate_airflow_env is not None:
            self.args.validate_airflow_env = validate_airflow_env

    @property
    def airflow_active(self) -> bool:
        return validate_env_vars(self.validate_airflow_env)

    ######################################################
    ## Build and validate data product
    ######################################################

    def build(self) -> bool:
        logger.debug(f"@build not defined for {self.__class__.__name__}")
        return False

    ######################################################
    ## Run data product
    ######################################################

    def run_in_local_env(self) -> bool:
        """
        Runs a DataProduct in the local environment where phi wf is called from.

        Returns:
            run_status (bool): True if the run was successful
        """
        logger.debug("--**-- Running DataProduct locally")

        wf_run_status: List[RunStatus] = []
        for idx, wf in enumerate(self.workflows, start=1):
            wf_name = wf.name or "{}__{}".format(wf.__class__.__name__, idx)
            print_subheading(f"\nRunning {wf_name}")
            # logger.debug("Setting context")
            # Pass down context
            wf.run_context = self.run_context
            wf.path_context = self.path_context
            # Not required for local run but adding for posterity
            if self.dag_id:
                wf.dag_id = self.dag_id
            run_success = wf.run_in_local_env()
            wf_run_status.append(RunStatus(wf_name, run_success))

        print_subheading("\nWorkflow run status:")
        print_info(
            "\n".join(
                [
                    "{}: {}".format(wf.name, "Success" if wf.success else "Fail")
                    for wf in wf_run_status
                ]
            )
        )
        print_info("")
        for _run in wf_run_status:
            if not _run.success:
                return False
        return True

    def run_in_docker_container(
        self, active_container: Any, docker_env: Optional[Dict[str, str]] = None
    ) -> bool:
        """
        Runs a DataProduct in a docker container.

        Args:
            active_container:
            docker_env:

        Returns:
            run_status (bool): True if the run was successful

        Notes:
            * This function runs in the local environment where phi wf is called from.
            But executes `airflow` commands in the docker container to run the workflow
            * For the airflow tasks to be available, they need to be added to the DAG
            using add_airflow_tasks_to_dag()
        """
        logger.debug("--**-- Running DataProduct in docker container")

        wf_run_status: List[RunStatus] = []
        for idx, wf in enumerate(self.workflows, start=1):
            wf_name = wf.name or "{}__{}".format(wf.__class__.__name__, idx)
            print_subheading(f"\nRunning {wf_name}")
            # Pass down context
            wf.run_context = self.run_context
            wf.path_context = self.path_context
            wf.dag_id = self.dag_id
            run_success = wf.run_in_docker_container(active_container, docker_env)
            wf_run_status.append(RunStatus(wf_name, run_success))

        print_subheading("\nWorkflow run status:")
        print_info(
            "\n".join(
                [
                    "{}: {}".format(wf.name, "Success" if wf.success else "Fail")
                    for wf in wf_run_status
                ]
            )
        )
        print_info("")
        for _run in wf_run_status:
            if not _run.success:
                return False
        return True

    def run_in_k8s_container(
        self,
        pod: Any,
        k8s_api_client: Any,
        container_name: Optional[str] = None,
        k8s_env: Optional[Dict[str, str]] = None,
    ) -> bool:
        """
        Runs a DataProduct in a k8s container.

        Args:
            pod: The pod to run the workflow in
            k8s_api_client: K8sApiClient
            container_name: The container to run the workflow in
            k8s_env:

        Returns:
            run_status (bool): True if the run was successful

        Notes:
            * This function runs in the local environment where phi wf is called from.
            But executes `airflow` commands in the k8s container to run the workflow
            * For the airflow tasks to be available, they need to be added to the DAG
            using add_airflow_tasks_to_dag()
        """
        logger.debug("--**-- Running DataProduct in k8s container")

        wf_run_status: List[RunStatus] = []
        for idx, wf in enumerate(self.workflows, start=1):
            wf_name = wf.name or "{}__{}".format(wf.__class__.__name__, idx)
            print_subheading(f"\nRunning {wf_name}")
            # Pass down context
            wf.run_context = self.run_context
            wf.path_context = self.path_context
            wf.dag_id = self.dag_id
            run_success = wf.run_in_k8s_container(
                pod=pod,
                k8s_api_client=k8s_api_client,
                container_name=container_name,
                k8s_env=k8s_env,
            )
            wf_run_status.append(RunStatus(wf_name, run_success))

        print_subheading("\nWorkflow run status:")
        print_info(
            "\n".join(
                [
                    "{}: {}".format(wf.name, "Success" if wf.success else "Fail")
                    for wf in wf_run_status
                ]
            )
        )
        print_info("")
        for _run in wf_run_status:
            if not _run.success:
                return False
        return True

    ######################################################
    ## Airflow functions
    ######################################################

    def create_airflow_dag(
        self,
        owner: Optional[str] = "airflow",
        depends_on_past: Optional[bool] = False,
        # The description for the DAG to e.g. be shown on the webserver
        description: Optional[str] = None,
        # Defines how often that DAG runs, this
        #  timedelta object gets added to your latest task instance's
        #  execution_date to figure out the next schedule
        # Default: daily
        schedule_interval: timedelta = timedelta(days=1),
        # The timestamp from which the scheduler will
        #  attempt to backfill
        start_date: Optional[datetime] = None,
        # if start_date is not provided, use start_days_ago
        start_days_ago: int = 2,
        # A date beyond which your DAG won't run, leave to None
        #  for open ended scheduling
        end_date: Optional[datetime] = None,
        # a dictionary of macros that will be exposed
        #  in your jinja templates. For example, passing ``dict(foo='bar')``
        #  to this argument allows you to ``{{ foo }}`` in all jinja
        #  templates related to this DAG. Note that you can pass any
        #  type of object here.
        user_defined_macros: Optional[Dict] = None,
        # a dictionary of filters that will be exposed
        #  in your jinja templates. For example, passing
        #  ``dict(hello=lambda name: 'Hello %s' % name)`` to this argument allows
        #  you to ``{{ 'world' | hello }}`` in all jinja templates related to
        #  this DAG.
        user_defined_filters: Optional[Dict] = None,
        # A dictionary of default parameters to be used
        #  as constructor keyword parameters when initialising operators.
        #  Note that operators have the same hook, and precede those defined
        #  here, meaning that if your dict contains `'depends_on_past': True`
        #  here and `'depends_on_past': False` in the operator's call
        #  `default_args`, the actual value will be `False`.
        default_args: Optional[Dict] = None,
        # the number of task instances allowed to run concurrently
        concurrency: Optional[int] = None,
        # maximum number of active DAG runs, beyond this
        #  number of DAG runs in a running state, the scheduler won't create
        #  new active DAG runs
        max_active_runs: int = 8,
        doc_md: Optional[str] = None,
        # a dictionary of DAG level parameters that are made
        # accessible in templates, namespaced under `params`. These
        # params can be overridden at the task level.
        params: Optional[Dict] = None,
        is_paused_upon_creation: Optional[bool] = None,
        jinja_environment_kwargs: Optional[Dict] = None,
        tags: Optional[List[str]] = None,
    ) -> Any:
        """
        Used to create an Airflow DAG for the DataProduct.
        """

        # Skips DAG creation on local machines
        if not self.airflow_active:
            return None

        from airflow import DAG
        from airflow.utils.dates import days_ago

        operator_default_args = default_args
        if operator_default_args is None:
            operator_default_args = {}
        operator_default_args.update(
            {
                "owner": owner,
                "depends_on_past": depends_on_past,
            }
        )
        if start_date is None:
            start_date = days_ago(start_days_ago)

        dag_id = self.dag_id
        if dag_id is None:
            print_error("DataProduct dag_id unavailable")
            return False
        dag = DAG(
            dag_id=dag_id,
            description=description,
            schedule_interval=schedule_interval,
            start_date=start_date,
            end_date=end_date,
            user_defined_macros=user_defined_macros,
            user_defined_filters=user_defined_filters,
            default_args=operator_default_args,
            concurrency=concurrency,
            max_active_runs=max_active_runs,
            doc_md=doc_md,
            params=params,
            is_paused_upon_creation=is_paused_upon_creation,
            jinja_environment_kwargs=jinja_environment_kwargs,
            tags=tags,
        )
        for workflow in self.workflows:
            _add_task_success = workflow.add_airflow_tasks_to_dag(dag)
            if not _add_task_success:
                print_error(f"Tasks for workflow {workflow.name} could not be added")
        logger.debug(f"Airflow dag {dag_id} created")
        return dag
